"""Representing a GasOil object"""

from typing import Optional

import numpy as np
import pandas as pd
from scipy.interpolate import PchipInterpolator

import pyscal
from pyscal.constants import EPSILON as epsilon
from pyscal.constants import MAX_EXPONENT_KR, SWINTEGERS
from pyscal.utils.relperm import crosspoint, estimate_diffjumppoint, truncate_zeroness
from pyscal.utils.string import comment_formatter, df2str

logger = pyscal.getLogger_pyscal(__name__)


class GasOil:
    """Object to represent two-phase properties for gas and oil.

    Parametrizations available for relative permeability:

     * Corey
     * LET

    or data can alternatively be read in from tabulated data
    (as Pandas DataFrame).

    No support (yet) to add capillary pressure.

    krgend is by default anchored both to `1-swl-sorg`, but can be set to
    anchor to `1-swl` instead. If the krgendanchor argument is something
    else than the string `sorg`, it will be anchored to `1-swl`.

    Arguments:
        swirr: Absolute minimal water saturation at infinite capillary
            pressure.  Not in use currently, except for in informational headers and
            for consistency checks.
        swl: First water saturation point in water tables. In GasOil, it
            is used to obtain the normalized oil and gas saturation.
        sgcr: Critical gas saturation. Gas will not be mobile before the
            gas saturation is above this value.
        sorg: Residual oil saturation after gas flooding. At this oil
            saturation, the oil has zero relative permeability.
        sgro: Residual gas, for use in gas-condensate modelling. Used
            as an endpoint for parametrized oil curve. Must be zero or equal
            to sgcr for compatibility with Eclipse three-point scaling.
        krgendanchor: Set to `sorg` (default) or something else, where to
            anchor `krgend`. If `sorg`, then the normalized gas
            saturation will be equal to 1 at `1 - swl - sorg`,
            if not, it will be 1 at `1 - swl`. If `sorg` is zero
            it does not matter. krgmax is only relevant when this anchor is sorg.
        h: Saturation step-length in the outputted table.
        tag: Optional string identifier, only used in comments.
        fast: Set to True if in order to skip some integrity checks
            and nice-to-have features. Not needed to set for normal pyscal
            runs, as speed is seldom crucial. Default False
    """

    def __init__(
        self,
        swirr: float = 0.0,
        sgcr: float = 0.0,
        h: Optional[float] = None,
        swl: float = 0.0,
        sorg: float = 0.0,
        sgro: float = 0.0,
        tag: str = "",
        krgendanchor: str = "sorg",
        fast: bool = False,
        _sgl: Optional[float] = None,  # Only to be used by GasWater.
    ) -> None:
        if h is None:
            h = 0.01

        assert -epsilon - 1 < swirr < 1.0 + epsilon, "-1 <= swirr <= 1 is required"
        assert -epsilon < sgcr < 1, "0 <= sgcr < 1 is required"
        assert -epsilon < swl < 1, "0 <= swl < 1 is required"
        assert -epsilon < sorg < 1, "0 <= sorg <  1 is required"

        if swirr < 0:
            logger.warning(
                f"Negative swirr value, {swirr}, detected. Negative values are allowed,"
                " but you should ensure that this is intentional."
            )

        if krgendanchor is None:
            krgendanchor = ""

        assert isinstance(krgendanchor, str), "krgendanchor must be a string"

        h_min = 10.0 / float(SWINTEGERS)
        if h < h_min:
            logger.warning(
                "Requested saturation step length (%g) too small, reset to %g", h, h_min
            )
            self.h = h_min
        else:
            self.h = h

        swl = max(swl, swirr)  # Can't allow swl < swirr, should we warn user?
        self.swl = swl
        self.swirr = swirr

        # Avoid endpoints close to zero, set them to zero if
        # they are below a certain limit:
        self.sorg = truncate_zeroness(sorg, name="sorg")
        self.sgcr = truncate_zeroness(sgcr, name="sgcr")
        self.sgro = truncate_zeroness(sgro, name="sgro")

        if not (np.isclose(self.sgro, 0) or np.isclose(self.sgro - self.sgcr, 0)):
            raise ValueError(
                "sgro must be zero or equal to sgcr, for compatibility with "
                "Eclipse three-point scaling. "
                f"sgro: {sgro}, sgcr: {sgcr}."
            )

        self.tag = tag

        if _sgl is not None:
            assert -epsilon < _sgl < sgcr + epsilon, "0 <= sgl <= sgcr is required"
            self.sgl = truncate_zeroness(_sgl, name="_sgl")
        else:
            self.sgl = 0.0

        if krgendanchor in ["sorg", ""]:
            self.krgendanchor = krgendanchor
        else:
            logger.warning("Unknown krgendanchor %s, ignored", str(krgendanchor))
            self.krgendanchor = ""

        self.fast = fast

        if np.isclose(self.sorg, 0.0) and self.krgendanchor == "sorg":
            self.krgendanchor = ""  # This is critical to avoid bugs due to numerics.

        if self.krgendanchor == "sorg" and not 1 - self.sorg - self.swl - self.sgcr > 0:
            raise ValueError(
                "No saturation range left for gas curve between endpoints, check input"
            )
        if self.krgendanchor == "" and not 1 - self.swl - self.sgcr > 0:
            raise ValueError(
                "No saturation range left for gas curve between endpoints, check input"
            )

        sg_list = [
            0.0,
            self.sgl,
            self.sgcr,
            *list(np.arange(self.sgcr + self.h, 1 - self.sorg - self.swl, self.h)),
            1 - self.sorg - self.swl,
            1 - self.swl,
        ]
        sg_list.sort()
        self.table = pd.DataFrame(sg_list, columns=["SG"])
        self.table["sgint"] = list(map(round, self.table["SG"] * SWINTEGERS))
        self.table = self.table.drop_duplicates("sgint")

        # Now sg=1-sorg-swl might be accidentally dropped, so make sure we
        # have it by replacing the closest value by 1 - sorg exactly
        sorgindex = (
            (self.table["SG"] - (1 - self.sorg - self.swl)).abs().sort_values().index[0]
        )
        self.table.loc[sorgindex, "SG"] = 1 - self.sorg - self.swl

        sgcrindex = (self.table["SG"] - (self.sgcr)).abs().sort_values().index[0]
        self.table.loc[sgcrindex, "SG"] = self.sgcr

        # Need to conserve sg=0 and sgcr:
        if sgcrindex == 0 and self.sgcr > 0.0:
            # This code is a safeguard againts truncate_zeroness(), which normally
            # prevents this from happening.
            zero_row = pd.DataFrame({"SG": 0}, index=[0])
            self.table = pd.concat([zero_row, self.table], sort=False).reset_index(
                drop=True
            )

        self.table = self.table.reset_index()
        self.table = self.table[["SG"]]
        self.table["SL"] = 1 - self.table["SG"]
        if krgendanchor == "sorg":
            # Normalized sg (sgn) is 0 at sgcr, and 1 at 1-swl-sorg
            assert 1 - swl - sgcr - sorg > epsilon
            self.table["SGN"] = (self.table["SG"] - self.sgcr) / (
                1 - self.swl - self.sgcr - self.sorg
            )
        else:
            assert 1 - swl - sgcr > epsilon
            self.table["SGN"] = (self.table["SG"] - self.sgcr) / (
                1 - self.swl - self.sgcr
            )

        # Normalized oil saturation should be 0 at sg=1-swl-sorg, and 1 at sg=sgro
        self.table["SON"] = (self.table["SL"] - self.sorg - self.swl) / (
            1 - self.sorg - self.swl - self.sgro
        )
        self.sgcomment: str = ""
        self.update_sgcomment_and_sorg()
        self.krgcomment = ""
        self.krogcomment = ""
        self.pccomment = ""

        logger.debug(
            "Initialized GasOil with %s saturation points", str(len(self.table))
        )

    def update_sgcomment_and_sorg(self):
        """Recalculate sorg in case it has table data has been manipulated"""
        self.sgcomment = (
            f"-- swirr={self.swirr:g}, sgcr={self.sgcr:g}, swl={self.swl:g}, "
            f"sorg={self.sorg:g}, sgro={self.sgro:g}, "
            f"krgendanchor={self.krgendanchor}\n"
        )
        if "KROG" in self.table.columns:
            self.sorg = (
                1
                - self.swl
                - self.table[np.isclose(self.table["KROG"], 0.0)].min()["SG"]
            )

    def add_fromtable(
        self,
        dframe: pd.DataFrame,
        sgcolname: str = "SG",
        krgcolname: str = "KRG",
        krogcolname: str = "KROG",
        pccolname: str = "PCOG",
        krgcomment: str = "",
        krogcomment: str = "",
        pccomment: str = "",
    ):
        """Interpolate relpermdata from a dataframe.

        The saturation range with endpoints must be set up beforehand,
        and must be compatible with the tabular input. The tabular
        input will be interpolated to the initialized Sg-table.

        If you have krg and krog in different dataframes, call this
        function twice

        Calling function is responsible for checking if any data was
        actually added to the table.
        """
        # Avoid having to deal with multi-indices:
        if len(dframe.index.names) > 1:
            logger.warning(
                "add_fromtable() did a reset_index(), consider not supplying MultiIndex"
            )
            dframe = dframe.reset_index()

        if sgcolname not in dframe:
            raise ValueError(
                f"{sgcolname} not found in dataframe, can't read table data"
            )

        for col in [sgcolname, krgcolname, krogcolname, pccolname]:
            # Typecheck/convert all numerical columns:
            if col in dframe and not pd.api.types.is_numeric_dtype(dframe[col]):
                # Try to convert to numeric type
                try:
                    dframe[col] = dframe[col].astype(float)
                    logger.info("Converted column %s to numbers for fromtable()", col)
                except (TypeError, ValueError) as err:
                    raise ValueError(
                        f"Failed to parse column {col} as numbers for add_fromtable()"
                    ) from err

        if dframe[sgcolname].min() > 0.0:
            raise ValueError("sg must start at zero")
        swlfrominput = 1 - dframe[sgcolname].max()
        if abs(swlfrominput - self.swl) > epsilon:
            logger.warning(
                "swl=%f and 1-max(sg)=%f from incoming table do not seem compatible",
                self.swl,
                swlfrominput,
            )
            logger.warning("         Do not trust the result near the endpoint.")

        if 0 < swlfrominput - self.swl < epsilon:
            # Perturb max sg in incoming dataframe when we are this close,
            # or we will get into floating trouble when interpolating.
            dframe.loc[dframe[sgcolname].idxmax(), sgcolname] += swlfrominput - self.swl

        if krgcolname in dframe:
            if not (dframe[krgcolname].diff().dropna() > -epsilon).all():
                raise ValueError("Incoming krg not increasing")
            if dframe[krgcolname].max() > 1.0:
                raise ValueError("krg is above 1 in incoming table")
            if dframe[krgcolname].min() < 0.0:
                raise ValueError("krg is below 0 in incoming table")
            pchip = PchipInterpolator(
                dframe[sgcolname].astype(float), dframe[krgcolname].astype(float)
            )
            # Do not extrapolate this data. We will bfill and ffill afterwards
            self.table["KRG"] = pchip(self.table["SG"], extrapolate=False)
            self.table["KRG"] = self.table["KRG"].ffill()
            self.table["KRG"] = self.table["KRG"].bfill()
            self.table["KRG"] = self.table["KRG"].clip(lower=0.0, upper=1.0)
            self.krgcomment = "-- krg from tabular input" + krgcomment + "\n"
            self.sgcr = self.estimate_sgcr()
        if krogcolname in dframe:
            if not (dframe[krogcolname].diff().dropna() < epsilon).all():
                raise ValueError("Incoming krog not decreasing")
            if dframe[krogcolname].max() > 1.0:
                raise ValueError("krog is above 1 in incoming table")
            if dframe[krogcolname].min() < 0.0:
                raise ValueError("krog is below 0 in incoming table")
            pchip = PchipInterpolator(
                dframe[sgcolname].astype(float), dframe[krogcolname].astype(float)
            )
            self.table["KROG"] = pchip(self.table["SG"], extrapolate=False)
            self.table["KROG"] = self.table["KROG"].ffill()
            self.table["KROG"] = self.table["KROG"].bfill()
            self.table["KROG"] = self.table["KROG"].clip(lower=0.0, upper=1.0)
            self.krogcomment = "-- krog from tabular input" + krogcomment + "\n"
            self.sorg = self.estimate_sorg()

            # For tabulated relperm data, correct knowlegde of sgro
            # is not important to pyscal (it is relevant when add_corey_gas()
            # etc is called). It is estimated here just for convenience.
            sgro_estimate = self.estimate_sgro()
            if not (
                np.isclose(sgro_estimate, 0.0) or np.isclose(sgro_estimate, self.sgcr)
            ):
                logger.warning(
                    "Estimated sgro (%s) from tabulated data was not 0 or sgcr (%s). "
                    "Reset to zero.",
                    str(sgro_estimate),
                    str(self.sgcr),
                )
                self.sgro = 0.0
            else:
                self.sgro = sgro_estimate
        if pccolname in dframe:
            # Incoming dataframe must cover the range:
            if dframe[sgcolname].max() < self.table["SG"].max():
                raise ValueError(
                    f"Too large swl for pcog interpolation, "
                    f"max incoming sg is {dframe[sgcolname].max()} "
                    f"and existing max(SG) is {self.table['SG'].max()}"
                )
            if np.isinf(dframe[pccolname]).any():
                logger.warning(
                    (
                        "Infinity pc values detected. Will be dropped, "
                        "risk of extrapolation"
                    )
                )
            dframe = dframe.replace([np.inf, -np.inf], np.nan)
            dframe = dframe.dropna(subset=[pccolname], how="all")
            # If nonzero, then it must be increasing:
            if (
                dframe[pccolname].abs().sum() > 0
                and not (dframe[pccolname].diff().dropna() > 0.0).all()
            ):
                raise ValueError("Incoming pc not increasing")
            pchip = PchipInterpolator(
                dframe[sgcolname].astype(float), dframe[pccolname].astype(float)
            )
            self.table["PC"] = pchip(self.table["SG"], extrapolate=False)
            if np.isnan(self.table["PC"]).any() or np.isinf(self.table["PC"]).any():
                raise ValueError("inf/nan in interpolated data, check input")
            self.pccomment = "-- pc from tabular input" + pccomment + "\n"

    def set_endpoints_linearpart_krg(
        self, krgend: float, krgmax: Optional[float] = None
    ):
        """Set linear parts of krg outside endpoints.

        Curve is set to zero in [0, sgcr].

        Given the default krgendanchor==sorg, the
        curve will be linear in [1 - swl - sorg, 1 - swl]
        (from krgend to krgmax). If not anchored to sorg, there
        is no linear part near sg=1-swl.

        If krgendanchor is set to `sorg` (default), then the normalized
        gas saturation `sgn` (which is what is raised to the power of `ng`)
        is 1 at `sg = 1 - swl - sorg`. If not, it is 1 at `sg = 1 - swl`.

        krgmax is only relevant if krgendanchor is 'sorg'

        This function is used by add_corey/LET_gas(), and perhaps by other
        utility functions. It should not be necessary for end-users.

        Args:
            krgend: krg at sg = 1 - swl - sorg.
            krgmax: krg at Sg = 1 - swl. Default 1.

        """
        self.table.loc[self.table["SG"] <= self.sgcr, "KRG"] = 0

        if self.krgendanchor == "sorg":
            # Linear curve between krgendcanchor and 1-swl if krgend
            # is anchored to sorg
            if not krgmax:
                krgmax = 1
            tmp = pd.DataFrame(self.table[["SG"]])
            tmp["sgendnorm"] = (tmp["SG"] - (1 - (self.sorg + self.swl))) / (self.sorg)
            tmp["KRG"] = (
                tmp["sgendnorm"] * krgmax + (1 - tmp["sgendnorm"]) * krgend
            ).clip(lower=0.0, upper=1.0)
            self.table.loc[
                self.table["SG"] >= (1 - (self.sorg + self.swl + epsilon)), "KRG"
            ] = tmp.loc[tmp["SG"] >= (1 - (self.sorg + self.swl + epsilon)), "KRG"]
        else:
            self.table.loc[self.table["SG"] > (1 - (self.swl + epsilon)), "KRG"] = (
                krgend
            )
            if krgmax and krgmax < 1.0 and self.sorg > 0:
                # Only warn if something else than default is in use
                logger.warning("krgmax ignored when not anchoring to sorg")

    def set_endpoints_linearpart_krog(
        self,
        kroend: float,
        kromax: Optional[float] = None,
    ):
        """Set linear parts of krog outside endpoints.

        Linear for sg in [0, sgro], from kromax to kroend, but nonzero
        sgro should only be used in gas-condensate modelling. When sgro
        is zero, kromax will be ignored.

        Zero for sg above 1 - sorg - swl.

        This function is used by add_corey/LET_oil(), and perhaps by other
        utility functions. It should not be necessary for end-users.

        Args:
            kroend: krog at sg=sgro, also if sgro=0
            kromax: krog at sg=0 for sgro > 0
        """
        if kromax is not None:
            if np.isclose(self.sgro, 0) and not np.isclose(kromax, kroend):
                logger.warning("kromax ignored when sgro is zero")
                kromax = kroend
            else:
                assert kroend <= kromax
        else:
            kromax = kroend

        # Special handling of the part close to sg=1, set to zero.
        self.table.loc[
            self.table["SG"] > 1 - self.sorg - self.swl - epsilon, "KROG"
        ] = 0

        # Floating point issues can cause a slight overshoot at sg=0:
        self.table.loc[self.table["KROG"] > kromax, "KROG"] = kromax

        self.table.loc[0, "KROG"] = kromax

    def add_corey_gas(
        self, ng: float = 2.0, krgend: float = 1.0, krgmax: Optional[float] = None
    ):
        """Add krg data through the Corey parametrization

        A column called 'krg' will be added. If it exists, it will
        be replaced.

        If krgendanchor is sorg, the Corey curve ends at krgend at
        sg = 1 - swl - sorg, and then linear up to krgmax at
        sg = 1 - swl. If not, it ends at krgend at sg = 1 - swl.

        krgmax is only relevant if krgendanchor is 'sorg'
        """
        assert epsilon < ng < MAX_EXPONENT_KR
        assert 0 < krgend <= 1.0
        if krgmax is not None:
            assert 0 < krgend <= krgmax <= 1.0
        self.table["KRG"] = krgend * self.table["SGN"] ** ng

        self.set_endpoints_linearpart_krg(krgend, krgmax)

        if not krgmax:
            krgmax = 1
        self.krgcomment = (
            f"-- Corey krg, ng={ng:g}, krgend={krgend:g}, krgmax={krgmax:g}\n"
        )

    def add_corey_oil(
        self,
        nog: float = 2,
        kroend: float = 1,
        kromax: Optional[float] = None,
    ):
        """
        Add kro data through the Corey parametrization

        A column named 'kro' will be added to the internal DataFrame,
        replaced if it exists.

        All values above 1 - sorg - swl are set to zero.

        Arguments:
            nog: Corey exponent for oil
            kroend: Value for krog at normalized oil saturation 1
            kromax: Value for sg=0 if sgro > 0.

        Returns:
            None (modifies internal class state)
        """
        assert epsilon < nog < MAX_EXPONENT_KR
        assert 0 < kroend <= 1.0

        self.table["KROG"] = kroend * self.table["SON"] ** nog

        self.set_endpoints_linearpart_krog(kroend, kromax)

        self.krogcomment = f"-- Corey krog, nog={nog:g}, kroend={kroend:g}"
        if kromax is not None:
            self.krogcomment += f", kromax={kromax:g}"
        self.krogcomment += "\n"

    def add_LET_gas(
        self,
        l: float = 2.0,
        e: float = 2.0,
        t: float = 2.0,
        krgend: float = 1.0,
        krgmax: Optional[float] = None,
    ):
        """
        Add gas relative permability data through the LET parametrization

        A column called 'KRG' will be added, replaced if it does not exist

        If krgendanchor is sorg, the LET curve ends at krgend at
        sg = 1 - swl - sorg, and then linear up to krgmax at
        sg = 1 - swl. If not, it ends at krgend at sg = 1 - swl.

        Arguments:
            l: L parameter in LET
            e: E parameter in LET
            t: T parameter in LET
            krgend: Value of krg at normalized gas saturation 1
            krgmax: Value of krg at gas saturation 1

        Returns:
            None (modifies internal state)
        """
        # Similar code in wateroil.add_LET_water, but readability
        # is better by having them separate
        assert epsilon < l < MAX_EXPONENT_KR
        assert epsilon < e < MAX_EXPONENT_KR
        assert epsilon < t < MAX_EXPONENT_KR
        if krgmax:
            assert 0 < krgend <= krgmax <= 1.0
        else:
            assert 0 < krgend <= 1.0

        self.table["KRG"] = (
            krgend
            * self.table["SGN"] ** l
            / ((self.table["SGN"] ** l) + e * (1 - self.table["SGN"]) ** t)
        )
        # This equation is undefined for t a float and sgn=1, set explicitly:
        self.table.loc[np.isclose(self.table["SGN"], 1.0), "KRG"] = krgend

        self.set_endpoints_linearpart_krg(krgend, krgmax)

        if not krgmax:
            krgmax = 1
        self.krgcomment = (
            f"-- LET krg, l={l:g}, e={e:g}, t={t:g}, "
            f"krgend={krgend:g}, krgmax={krgmax:g}\n"
        )

    def add_LET_oil(
        self,
        l: float = 2.0,
        e: float = 2.0,
        t: float = 2.0,
        kroend: float = 1.0,
        kromax: Optional[float] = None,
    ):
        """Add oil (vs gas) relative permeability data through the Corey
        parametrization.

        A column named 'krog' will be added, replaced if it exists.

        All values where sg > 1 - sorg - swl are set to zero.

        Arguments:
            l: L parameter
            e: E parameter
            t: T parameter
            kroend: The value at gas saturation sgcr
            kromax: Value at sg=0 for sgro > 0

        """
        assert epsilon < l < MAX_EXPONENT_KR
        assert epsilon < e < MAX_EXPONENT_KR
        assert epsilon < t < MAX_EXPONENT_KR
        assert 0 < kroend <= 1.0

        self.table["KROG"] = (
            kroend
            * self.table["SON"] ** l
            / ((self.table["SON"] ** l) + e * (1 - self.table["SON"]) ** t)
        )
        # This equation is undefined for t a float and son=1, set explicitly:
        self.table.loc[np.isclose(self.table["SON"], 1.0), "KROG"] = kroend

        self.set_endpoints_linearpart_krog(kroend, kromax)

        self.krogcomment = f"-- LET krog, l={l:g}, e={e:g}, t={t:g}, kroend={kroend:g}"
        if kromax is not None:
            self.krogcomment += f", kromax={kromax:g}"
        self.krogcomment += "\n"

    def estimate_sgro(self):
        """Estimate sgro of the current krog data

        sgro is estimated by searching for a linear part in kro
        from sg=0. In practice it is impossible to infer sgro = 0,
        since we are limited by h.

        When initializing GasOil, sgro must be either zero or equal
        to sgcr. This function only estimates sgro, and will not guarantee
        that condition.

        If the curve is linear everywhere, sgro will be returned as 1 - swl + h

        Returns:
            float: The estimated sgro
        """
        assert "KROG" in self.table
        assert self.table["KROG"].sum() > 0
        return estimate_diffjumppoint(self.table, xcol="SG", ycol="KROG", side="left")

    def estimate_sorg(self) -> float:
        """Estimate sorg of the current krg or krog data.

        sorg is estimated by searching for a linear part in krg downwards
        from sg=1-swl. In practice it is impossible to infer sorg = 0,
        since we are limited by h, and the last segment from sg=1-swl-h
        to sg=1-swl must always be assumed linear.

        If the curve is linear everywhere, sorg will be returned as sgcr + h

        If krgend is anchored to sorg, krg data is used to infer sorg. If not,
        krg cannot be used for this, and krog is used. sorg might be overestimated
        when krog is used if it very close to zero before reaching sorg.

        Returns:
            The estimated sorg.
        """
        if self.krgendanchor == "sorg":
            assert "KRG" in self.table
            assert self.table["KRG"].sum() > 0
            return self.table["SG"].max() - estimate_diffjumppoint(
                self.table, xcol="SG", ycol="KRG", side="right"
            )
        assert "KROG" in self.table
        assert self.table["KROG"].sum() > 0
        return self.table["SG"].max() - estimate_diffjumppoint(
            self.table, xcol="SG", ycol="KROG", side="right"
        )

    def estimate_sgcr(self) -> float:
        """Estimate sgcr of the current krog data.

        sgcr is the largest gas saturation for which the gas relative
        permeability is approximately zero, where approximately zero is
        roughly equivalent to how many digits are outputted
        by SGOF().

        Returns:
            The estimated sgcr.
        """
        return self.table[self.table["KRG"] < 10 * epsilon]["SG"].max()

    def crosspoint(self) -> float:
        """Locate and return the saturation point where krg = krog

        Accuracy of this crosspoint depends on the resolution chosen
        when initializing the saturation range (it uses linear
        interpolation to solve for the zero)

        Returns:
            The gas saturation where krg == krog, for relperm
            linearly interpolated in gas saturation.
        """
        return crosspoint(self.table, "SG", "KRG", "KROG")

    def selfcheck(self, mode: str = "SGOF") -> bool:
        """Check validities of the data in the table.

        This is to catch errors that are either physically wrong
        or at least causes Eclipse 100 to stop.

        Returns True if no errors are found, False if at least one
        is found.

        If you call SGOF/SLGOF, this function must not return False.

        Args:
            mode: If mode is "SGFN", krog is not required.
        """
        error = False
        if "KRG" not in self.table:
            logger.error("KRG data missing")
            error = True
        if not (self.table["SG"].diff().dropna() > -epsilon).all():
            logger.error("SG data not strictly increasing")
            error = True
        if (
            "KRG" in self.table
            and not (self.table["KRG"].diff().dropna() >= -epsilon).all()
        ):
            logger.error("KRG data not monotonically decreasing")
            error = True

        if mode != "SGFN":
            if "KROG" not in self.table:
                logger.error("KROG data missing")
                error = True
            if (
                "KROG" in self.table
                and not (self.table["KROG"].diff().dropna() <= epsilon).all()
            ):
                logger.error("KROG data not monotonically increasing")
                error = True
        if "KRG" in self.table and not np.isclose(min(self.table["KRG"]), 0.0):
            logger.error("KRG must start at zero")
            error = True
        if (
            "PC" in self.table
            and self.table["PC"][0] > -epsilon
            and not (self.table["PC"].diff().dropna() < epsilon).all()
        ):
            logger.error("PC data for gas-oil not strictly decreasing")
            error = True
        if "PC" in self.table and np.isinf(self.table["PC"].max()):
            logger.error("PC goes to infinity for gas-oil. ")
            error = True
        if "PC" in self.table.columns and np.isnan(self.table["PC"]).any():
            logger.error("pc data contains NaN")
            error = True

        for col in list({"SG", "KRG", "KROG"} & set(self.table.columns)):
            if not (
                (min(self.table[col]) >= -epsilon)
                and (max(self.table[col]) <= 1 + epsilon)
            ):
                logger.error("%s data should be contained in [0,1]", col)
                error = True
        if error:
            return False
        logger.debug("GasOil object is checked to be valid")
        return True

    def SGOF(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """
        Produce SGOF input for Eclipse reservoir simulator.

        The columns sg, krg, krog and pc are outputted and
        formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: Whether the SGOF string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGOF yourself).
                Defaults to True.
            dataincommentrow: Whether metadata should be printed,
                defaults to True.
        """
        if not self.fast and not self.selfcheck():
            # selfcheck() will log error/warning messages
            return ""
        string = ""
        if "PC" not in self.table:
            self.table["PC"] = 0.0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SGOF\n"
        string += comment_formatter(self.tag)
        string += "-- pyscal: " + str(pyscal.__version__) + "\n"
        if dataincommentrow:
            string += self.sgcomment
            string += self.krgcomment
            string += self.krogcomment
            if not self.fast:
                string += f"-- krg = krog @ sg={self.crosspoint():1.5f}\n"
            string += self.pccomment
        width = 10
        string += (
            "-- "
            + "SG".ljust(width - 3)
            + "KRG".ljust(width)
            + "KROG".ljust(width)
            + "PC".ljust(width)
            + "\n"
        )
        string += df2str(
            self.table[["SG", "KRG", "KROG", "PC"]],
            monotonicity={
                "KROG": {"sign": -1, "lower": 0, "upper": 1},
                "KRG": {"sign": 1, "lower": 0, "upper": 1},
                "PC": {"sign": 1, "allowzero": True},
            }
            if not self.fast
            else None,
        )
        string += "/\n"
        return string

    def slgof_df(self) -> pd.DataFrame:
        """Slice out an SLGOF table.

        This is used by the SLGOF() function, it is
        extracted as a single function to facilitate testing."""
        if "PC" not in self.table.columns:
            # Only happens when the SLGOF function is skipped (test code)
            self.table["PC"] = 0.0
        return (
            self.table[
                self.table["SG"] <= 1 - self.sorg - self.swl + 1.0 / float(SWINTEGERS)
            ]
            .sort_values("SL")[["SL", "KRG", "KROG", "PC"]]
            .reset_index(drop=True)
        )

    def SLGOF(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Produce SLGOF input for Eclipse reservoir simulator.

        The columns sl (liquid saturation), krg, krog and pc are
        outputted and formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SLGOF string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGOF yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        if not self.selfcheck():
            # Selfcheck will issue error messages.
            return ""
        string = ""
        if "PC" not in self.table:
            self.table["PC"] = 0.0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SLGOF\n"
        string += comment_formatter(self.tag)
        string += "-- pyscal: " + str(pyscal.__version__) + "\n"
        if dataincommentrow:
            string += self.sgcomment
            string += self.krgcomment
            string += self.krogcomment
            string += f"-- krg = krog @ sg={self.crosspoint():1.5f}\n"
            string += self.pccomment
        width = 10
        string += (
            "-- "
            + "SL".ljust(width - 3)
            + "KRG".ljust(width)
            + "KROG".ljust(width)
            + "PC".ljust(width)
            + "\n"
        )
        string += df2str(
            self.slgof_df(),
            monotonicity={
                "KROG": {"sign": 1, "lower": 0, "upper": 1},
                "KRG": {"sign": -1, "lower": 0, "upper": 1},
                "PC": {"sign": -1, "allowzero": True},
            }
            if not self.fast
            else None,
        )
        string += "/\n"
        return string

    def SGFN(
        self,
        header: bool = True,
        dataincommentrow: bool = True,
        sgcomment: Optional[str] = None,
        crosspointcomment: Optional[str] = None,
    ):
        """
        Produce SGFN input for Eclipse reservoir simulator.

        The columns sg, krg, and pc are outputted and
        formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SGFN string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGFN yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
            sgcomment: Provide the string to include in the comment
                section for describing the saturation endpoints. Used
                by GasWater.
            crosspointcomment: String to be used for crosspoint comment
                string, overrides what this object can provide. Used by GasWater.
                If None, it will be computed, use empty string to avoid.
        """
        if not self.selfcheck(mode="SGFN"):
            # Selfcheck will issue error messages.
            return ""
        string = ""
        if "PC" not in self.table.columns:
            self.table["PC"] = 0.0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SGFN\n"
        string += comment_formatter(self.tag)
        string += "-- pyscal: " + str(pyscal.__version__) + "\n"
        if dataincommentrow:
            if sgcomment is not None:
                string += sgcomment
            else:
                string += self.sgcomment
            string += self.krgcomment
            if crosspointcomment is None:
                if "KROG" in self.table.columns:
                    string += f"-- krg = krog @ sg={self.crosspoint():1.5f}\n"
            else:
                string += crosspointcomment
            string += self.pccomment
        width = 10
        string += (
            "-- "
            + "SG".ljust(width - 3)
            + "KRG".ljust(width)
            + "PC".ljust(width)
            + "\n"
        )
        string += df2str(
            self.table[["SG", "KRG", "PC"]],
            monotonicity={
                "KRG": {"sign": 1, "lower": 0, "upper": 1},
                "PC": {"sign": 1, "allowzero": True},
            }
            if not self.fast
            else None,
        )
        string += "/\n"
        return string

    def GOTABLE(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """
        Produce GOTABLE input for the Nexus reservoir simulator.

        The columns sg, krg, krog and pc are outputted and
        formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SGOF string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGOF yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        string = ""
        if "PC" not in self.table.columns:
            self.table["PC"] = 0.0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "GOTABLE\n"
            string += "SG KRG KROG PC\n"
        string += "! pyscal: " + str(pyscal.__version__) + "\n"
        if dataincommentrow:
            string += self.sgcomment.replace("--", "!")
            string += self.krgcomment.replace("--", "!")
            string += self.krogcomment.replace("--", "!")
            string += f"! krg = krog @ sw={self.crosspoint():1.5f}\n"
            string += self.pccomment.replace("--", "!")
        width = 10
        string += (
            "! "
            + "SG".ljust(width - 2)
            + "KRG".ljust(width)
            + "KROG".ljust(width)
            + "PC".ljust(width)
            + "\n"
        )
        string += df2str(
            self.table[["SG", "KRG", "KROG", "PC"]],
            monotonicity={
                "KROG": {"sign": -1, "lower": 0, "upper": 1},
                "KRG": {"sign": 1, "lower": 0, "upper": 1},
                "PC": {"sign": 1, "allowzero": True},
            }
            if self.fast
            else None,
        )
        return string

    def plotkrgkrog(
        self,
        mpl_ax=None,
        color: str = "blue",
        alpha: float = 1.0,
        linewidth: int = 1,
        linestyle: str = "-",
        marker: Optional[str] = None,
        label: Optional[str] = None,
        logyscale: bool = False,
    ):
        """Plot krg and krog

        If mpl_ax is not None, it will be used as a
        matplotlib axis to plot on, if None, a fresh plot
        will be made.
        """
        # Lazy import of matplotlib for speed reasons
        import matplotlib  # noqa: PLC0415
        import matplotlib.pyplot as plt  # noqa: PLC0415

        if mpl_ax is None:
            matplotlib.style.use("ggplot")
            _, useax = matplotlib.pyplot.subplots()
        else:
            useax = mpl_ax
        if logyscale:
            useax.set_yscale("log")
            useax.set_ylim((1e-8, 1))
        self.table.plot(
            ax=useax,
            x="SG",
            y="KRG",
            c=color,
            alpha=alpha,
            legend=None,
            label=label,
            linewidth=linewidth,
            linestyle=linestyle,
            marker=marker,
        )
        self.table.plot(
            ax=useax,
            x="SG",
            y="KROG",
            c=color,
            alpha=alpha,
            legend=None,
            label=None,
            linewidth=linewidth,
            linestyle=linestyle,
            marker=marker,
        )
        if mpl_ax is None:
            plt.show()



"""pyscal"""

import logging
import sys
from typing import Dict, Optional, Union

try:
    from .version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"


def getLogger_pyscal(
    module_name: str = "pyscal", args_dict: Optional[Dict[str, Union[str, bool]]] = None
) -> logging.Logger:
    # pylint: disable=invalid-name
    """Provide a custom logger for pyscal
    Logging output is by default split by logging levels (split between WARNING and
    ERROR) to stdout and stderr, each log occurs in only one of the streams.
    Args:
        module_name: A suggested name for the logger, usually __name__ should be
        supplied
        args_dict: Dictionary with contents from the argparse namespace object
        containing only keys "output", "verbose" and "debug".
    """
    logger = logging.getLogger(module_name)
    if len(logger.handlers) != 0:
        return logger

    if args_dict is None:
        args_dict = {}
    formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")

    if args_dict.get("output") == "-":
        # If main output is to stdout, we must send all logs to stderr:
        default_handler = logging.StreamHandler(sys.stderr)
        default_handler.setFormatter(formatter)
        logger.addHandler(default_handler)
    else:
        # Split log messages to either stdout or stderr based on log level:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.addFilter(lambda record: record.levelno < logging.ERROR)
        stdout_handler.setFormatter(formatter)

        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.addFilter(lambda record: record.levelno >= logging.ERROR)
        stderr_handler.setFormatter(formatter)

        logger.addHandler(stdout_handler)
        logger.addHandler(stderr_handler)

    # --debug overrides --verbose
    if args_dict.get("debug", False):
        logger.setLevel(logging.DEBUG)
    elif args_dict.get("verbose", False):
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    if module_name == "pyscal.pyscalcli":
        all_modules = [
            "factory",
            "gasoil",
            "gaswater",
            "pyscallist",
            "scalrecommendation",
            "wateroil",
            "wateroilgas",
        ]
        for module in all_modules:
            module_logger = logging.getLogger("pyscal." + module)
            module_logger.handlers = []
            for handler in logger.handlers:
                module_logger.addHandler(handler)
            module_logger.setLevel(logger.level)

    return logger


# The order of imports must be conserved to avoid circular imports:
from .wateroil import WaterOil  # noqa
from .wateroilgas import WaterOilGas  # noqa
from .gasoil import GasOil  # noqa
from .gaswater import GasWater  # noqa
from .scalrecommendation import SCALrecommendation  # noqa
from .pyscallist import PyscalList  # noqa
from .factory import PyscalFactory  # noqa



"""Factory functions for creating the pyscal objects"""

import csv
import warnings
import zipfile
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Set, Union

import numpy as np
import openpyxl
import pandas as pd
import xlrd

from pyscal import getLogger_pyscal
from pyscal.utils import capillarypressure

from .gasoil import GasOil
from .gaswater import GasWater
from .pyscallist import PyscalList
from .scalrecommendation import SCALrecommendation
from .wateroil import WaterOil
from .wateroilgas import WaterOilGas

logger = getLogger_pyscal(__name__)


def slicedict(dct: dict, keys: Iterable):
    """Slice a dictionary for a set of keys.
    Keys not existing will be ignored.
    """
    return {k: dct[k] for k in keys if k in dct}


# Sets of dictionary keys, presence of these
# in incoming dictionary determines the codepaths. These must
# again match the API of the functions downstream in e.g. WaterOil, except
# for LET parameters, where the API is simplified to 'l', 'e' and 't'.
# We are case *insensitive* in this factory class, so everything here should
# be lower case
WO_INIT = ["swirr", "swl", "swcr", "sorw", "socr", "sgrw", "h", "tag"]
WO_COREY_WATER = ["nw"]
WO_WATER_ENDPOINTS = ["krwmax", "krwend"]
WO_COREY_OIL = ["now"]
WO_LET_WATER = ["lw", "ew", "tw"]  # Will translated to l, e and t in code below.
WO_LET_OIL = ["low", "eow", "tow"]
WO_LET_OIL_ALT = ["lo", "eo", "to"]  # Alternative parameter names.
WO_OIL_ENDPOINTS = ["kroend"]
WO_SIMPLE_J = ["a", "b", "poro_ref", "perm_ref", "drho"]  # "g" is optional
WO_SWLHEIGHT = ["swlheight"]
WO_SWCR_ADD = ["swcr_add"]  # Relevant when swlheight is in use.
WO_SWL_FROM_HEIGHT = [*WO_SWLHEIGHT, "swirr", "a", "b", "poro_ref", "perm_ref"]
WO_NORM_J = ["a", "b", "poro", "perm", "sigma_costau"]
# 'a' in WO_NORM_J is the same as a_petro, but should possibly kept as is.
WO_SIMPLE_J_PETRO = [
    "a_petro",
    "b_petro",
    "poro_ref",
    "perm_ref",
    "drho",
]  # "g" is optional
WO_LET_PC_PD = ["lpow", "epow", "tpow", "ltow", "etow", "ttow", "pcowmax", "pcowt"]
WO_LET_PC_IMB = [
    "lsow",
    "esow",
    "tsow",
    "lfow",
    "efow",
    "tfow",
    "pcowmax",
    "pcowmin",
    "pcowt",
]
WO_SKJAEVELAND_PC = ["cw", "co", "aw", "ao"]

GO_INIT = ["swirr", "sgcr", "sorg", "sgro", "swl", "krgendanchor", "h", "tag"]
GO_COREY_GAS = ["ng"]
GO_GAS_ENDPOINTS = ["krgend", "krgmax"]
GO_COREY_OIL = ["nog"]
GO_OIL_ENDPOINTS = [
    "kroend",
    "kromax",
]

GO_LET_GAS = ["lg", "eg", "tg"]
GO_LET_OIL = ["log", "eog", "tog"]
GO_LET_PC_PD = ["lpog", "epog", "tpog", "ltog", "etog", "ttog", "pcogmax", "pcogt"]
GO_LET_PC_IMB = [
    "lsog",
    "esog",
    "tsog",
    "lfog",
    "efog",
    "tfog",
    "pcogmax",
    "pcogmin",
    "pcogt",
]

GW_INIT = ["swirr", "swl", "sgl", "swcr", "sgrw", "sgcr", "h", "tag"]
GW_COREY_WATER = ["nw"]
GW_COREY_GAS = ["ng"]
GW_WATER_ENDPOINTS = ["krwmax", "krwend"]
GW_LET_WATER = ["lw", "ew", "tw"]
GW_LET_GAS = ["lg", "eg", "tg"]
GW_GAS_ENDPOINTS = ["krgmax", "krgend"]
GW_SIMPLE_J = ["a", "b", "poro_ref", "perm_ref", "drho"]  # "g" is optional
GW_NORM_J = ["a", "b", "poro", "perm", "sigma_costau"]
# 'a' in WO_NORM_J is the same as a_petro, but should possibly kept as is.
GW_SIMPLE_J_PETRO = [
    "a_petro",
    "b_petro",
    "poro_ref",
    "perm_ref",
    "drho",
]  # "g" is optional
GW_LET_PC_PD = ["lpgw", "epgw", "tpgw", "ltgw", "etgw", "ttgw", "pcgwmax", "pcgwt"]
GW_LET_PC_IMB = [
    "lsgw",
    "esgw",
    "tsgw",
    "lfgw",
    "efgw",
    "tfgw",
    "pcgwmax",
    "pcgwmin",
    "pcgwt",
]

WOG_INIT = ["swirr", "swl", "swcr", "sorw", "socr", "sorg", "sgcr", "h", "tag"]


def create_water_oil(
    params: Optional[Dict[str, float]] = None, fast: bool = False
) -> WaterOil:
    """Create a WaterOil object from a dictionary of parameters.

    Parameterization (Corey/LET) is inferred from presence
    of certain parameters in the dictionary.

    Don't rely on behaviour of you supply both Corey and LET at
    the same time.

    Parameter names in the dictionary are case insensitive. You
    can use Swirr, swirr, sWirR, swiRR etc.

    NB: the add_LET_* methods have the names 'l', 'e' and 't'
    in their signatures, which is not precise enough in this
    context, so we require e.g. 'Lw' and 'Low' (which both will be
    translated to 'l')

    Recognized parameters:
        swirr, swl, swcr, sorw, socr, sgrw, h, tag, nw, now, krwmax, krwend,
        lw, ew, tw, low, eow, tow, lo, eo, to, kroend,
        a, a_petro, b, b_petro, poro_ref, perm_ref, drho,
        a, b, poro, perm, sigma_costau

    Args:
        params: Dictionary with parameters describing the WaterOil object.
        fast: If fast-mode should be set for constructed object.
    """
    if not params:
        params = {}
    if not isinstance(params, dict):
        raise TypeError("Parameter to create_water_oil must be a dictionary")

    check_deprecated(params)

    sufficient_water_oil_params(params, failhard=True)

    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    # Allowing sending in NaN values, delete those keys.
    params = filter_nan_from_dict(params)

    usedparams: Set[str] = set()

    # Check if we should initialize swl from a swlheight parameter:
    if set(WO_SWL_FROM_HEIGHT).issubset(params):
        if "swl" in params:
            raise ValueError("Do not provide both swl and swlheight at the same time")
        if params["swlheight"] <= 0:
            raise ValueError("swlheight must be larger than zero")
        params_swl_from_height = slicedict(params, WO_SWL_FROM_HEIGHT)
        params["swl"] = capillarypressure.swl_from_height_simpleJ(
            **params_swl_from_height
        )
        logger.debug("Computed swl from swlwheight to %s", str(params["swl"]))
        if "swcr" in params and params["swcr"] < params["swl"]:
            raise ValueError(
                f"Provided swcr={params['swcr']} is lower than "
                f"computed swl={params['swl']}"
            )
    elif set(WO_SWLHEIGHT).issubset(params):
        raise ValueError(
            (
                "Can't initialize from SWLHEIGHT without sufficient "
                f"simple-J parameters, needs all of: {WO_SWL_FROM_HEIGHT}"
            )
        )

    # Should we have a swcr relative to swl?
    if set(WO_SWCR_ADD).issubset(params):
        if "swl" not in params:
            raise ValueError(
                (
                    "If swcr should be relative to swl, "
                    "both swcr_add and swl must be provided"
                )
            )
        if "swcr" in params:
            raise ValueError("Do not provide both swcr and swcr_add at the same time")
        params["swcr"] = params["swl"] + params[WO_SWCR_ADD[0]]

    # No requirements to the base objects, defaults are ok.
    wateroil = WaterOil(**alias_sgrw(slicedict(params, WO_INIT)), fast=fast)
    usedparams = usedparams.union(set(slicedict(params, WO_INIT).keys()))
    logger.debug(
        "Initialized WaterOil object from parameters %s", str(list(usedparams))
    )

    # Water curve
    params_corey_water = slicedict(params, WO_COREY_WATER + WO_WATER_ENDPOINTS)
    params_let_water = slicedict(params, WO_LET_WATER + WO_WATER_ENDPOINTS)
    if set(WO_COREY_WATER).issubset(set(params_corey_water)):
        wateroil.add_corey_water(**params_corey_water)
        logger.debug(
            "Added Corey water to WaterOil object from parameters %s",
            str(params_corey_water.keys()),
        )
    elif set(WO_LET_WATER).issubset(set(params_let_water)):
        params_let_water["l"] = params_let_water.pop("lw")
        params_let_water["e"] = params_let_water.pop("ew")
        params_let_water["t"] = params_let_water.pop("tw")
        wateroil.add_LET_water(**params_let_water)
        logger.debug(
            "Added LET water to WaterOil object from parameters %s",
            str(params_let_water.keys()),
        )

    # Oil curve:
    params_corey_oil = slicedict(params, WO_COREY_OIL + WO_OIL_ENDPOINTS)
    params_let_oil = slicedict(params, WO_LET_OIL + WO_LET_OIL_ALT + WO_OIL_ENDPOINTS)
    if set(WO_COREY_OIL).issubset(set(params_corey_oil)):
        wateroil.add_corey_oil(**params_corey_oil)
        logger.debug(
            "Added Corey water to WaterOil object from parameters %s",
            str(params_corey_oil.keys()),
        )
    elif set(WO_LET_OIL).issubset(set(params_let_oil)):
        params_let_oil["l"] = params_let_oil.pop("low")
        params_let_oil["e"] = params_let_oil.pop("eow")
        params_let_oil["t"] = params_let_oil.pop("tow")
        wateroil.add_LET_oil(**params_let_oil)
        logger.debug(
            "Added LET water to WaterOil object from parameters %s",
            str(params_let_oil.keys()),
        )
    elif set(WO_LET_OIL_ALT).issubset(set(params_let_oil)):
        params_let_oil["l"] = params_let_oil.pop("lo")
        params_let_oil["e"] = params_let_oil.pop("eo")
        params_let_oil["t"] = params_let_oil.pop("to")
        wateroil.add_LET_oil(**params_let_oil)
        logger.debug(
            "Added LET water to WaterOil object from parameters %s",
            str(params_let_oil.keys()),
        )

    # Capillary pressure:
    params_simple_j = slicedict(params, [*WO_SIMPLE_J, "g"])
    params_norm_j = slicedict(params, WO_NORM_J)
    params_simple_j_petro = slicedict(params, [*WO_SIMPLE_J_PETRO, "g"])
    params_let_pc_pd = slicedict(params, WO_LET_PC_PD)
    params_let_pc_imb = slicedict(params, WO_LET_PC_IMB)
    params_skjaeveland_pc = slicedict(params, WO_SKJAEVELAND_PC)

    if set(WO_SIMPLE_J).issubset(set(params_simple_j)):
        wateroil.add_simple_J(**params_simple_j)
    elif set(WO_SIMPLE_J_PETRO).issubset(set(params_simple_j_petro)):
        params_simple_j_petro["a"] = params_simple_j_petro.pop("a_petro")
        params_simple_j_petro["b"] = params_simple_j_petro.pop("b_petro")
        wateroil.add_simple_J_petro(**params_simple_j_petro)
    elif set(WO_NORM_J).issubset(set(params_norm_j)):
        wateroil.add_normalized_J(**params_norm_j)
    elif set(WO_LET_PC_PD).issubset(set(params_let_pc_pd)):
        params_let_pc_pd["Lp"] = params_let_pc_pd.pop("lpow")
        params_let_pc_pd["Ep"] = params_let_pc_pd.pop("epow")
        params_let_pc_pd["Tp"] = params_let_pc_pd.pop("tpow")
        params_let_pc_pd["Lt"] = params_let_pc_pd.pop("ltow")
        params_let_pc_pd["Et"] = params_let_pc_pd.pop("etow")
        params_let_pc_pd["Tt"] = params_let_pc_pd.pop("ttow")
        params_let_pc_pd["Pcmax"] = params_let_pc_pd.pop("pcowmax")
        params_let_pc_pd["Pct"] = params_let_pc_pd.pop("pcowt")
        wateroil.add_LET_pc_pd(**params_let_pc_pd)
    elif set(WO_LET_PC_IMB).issubset(set(params_let_pc_imb)):
        params_let_pc_imb["Ls"] = params_let_pc_imb.pop("lsow")
        params_let_pc_imb["Es"] = params_let_pc_imb.pop("esow")
        params_let_pc_imb["Ts"] = params_let_pc_imb.pop("tsow")
        params_let_pc_imb["Lf"] = params_let_pc_imb.pop("lfow")
        params_let_pc_imb["Ef"] = params_let_pc_imb.pop("efow")
        params_let_pc_imb["Tf"] = params_let_pc_imb.pop("tfow")
        params_let_pc_imb["Pcmax"] = params_let_pc_imb.pop("pcowmax")
        params_let_pc_imb["Pct"] = params_let_pc_imb.pop("pcowt")
        params_let_pc_imb["Pcmin"] = params_let_pc_imb.pop("pcowmin")
        wateroil.add_LET_pc_imb(**params_let_pc_imb)
    elif set(WO_SKJAEVELAND_PC).issubset(set(params_skjaeveland_pc)):
        wateroil.add_skjaeveland_pc(**params_skjaeveland_pc)
    else:
        logger.info(
            (
                "Missing or ambiguous parameters for capillary pressure in "
                "WaterOil object. Using zero."
            )
        )
    if not wateroil.selfcheck():
        raise ValueError(
            ("Incomplete WaterOil object, some parameters missing to factory")
        )
    return wateroil


def create_gas_oil(
    params: Optional[Dict[str, float]] = None, fast: bool = False
) -> GasOil:
    """Create a GasOil object from a dictionary of parameters.

    Parameterization (Corey/LET) is inferred from presence
    of certain parameters in the dictionary.

    Don't rely on behaviour of you supply both Corey and LET at
    the same time.

    NB: the add_LET_* methods have the names 'l', 'e' and 't'
    in their signatures, which is not precise enough in this
    context, so we require e.g. 'Lg' and 'Log' (which both will be
    translated to 'l').

    Recognized parameters:
        swirr, sgcr, sorg, swl, krgendanchor, h, tag,
        ng, krgend, krgmax, nog, kroend, kromax
        lg, eg, tg, log, eog, tog

    Args:
        params: Dictionary with parameters describing the GasOil object.
        fast: If fast-mode should be set for constructed object.
    """
    if not params:
        params = {}
    if not isinstance(params, dict):
        raise TypeError("Parameter to create_gas_oil must be a dictionary")

    check_deprecated(params)

    sufficient_gas_oil_params(params, failhard=True)

    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    # Allowing sending in NaN values, delete those keys.
    params = filter_nan_from_dict(params)

    usedparams: Set[str] = set()
    # No requirements to the base objects, defaults are ok.
    gasoil = GasOil(**slicedict(params, GO_INIT), fast=fast)
    usedparams = usedparams.union(set(slicedict(params, GO_INIT).keys()))
    logger.debug("Initialized GasOil object from parameters %s", str(list(usedparams)))

    # Gas curve
    params_corey_gas = slicedict(params, GO_COREY_GAS + GO_GAS_ENDPOINTS)
    params_let_gas = slicedict(params, GO_LET_GAS + GO_GAS_ENDPOINTS)
    if set(GO_COREY_GAS).issubset(set(params_corey_gas)):
        gasoil.add_corey_gas(**params_corey_gas)
        logger.debug(
            "Added Corey gas to GasOil object from parameters %s",
            str(params_corey_gas.keys()),
        )
    elif set(GO_LET_GAS).issubset(set(params_let_gas)):
        params_let_gas["l"] = params_let_gas.pop("lg")
        params_let_gas["e"] = params_let_gas.pop("eg")
        params_let_gas["t"] = params_let_gas.pop("tg")
        gasoil.add_LET_gas(**params_let_gas)
        logger.debug(
            "Added LET gas to GasOil object from parameters %s",
            str(params_let_gas.keys()),
        )
    else:
        logger.warning("Missing or ambiguous parameters for gas curve in GasOil object")

    # Oil curve:
    params_corey_oil = slicedict(params, GO_COREY_OIL + GO_OIL_ENDPOINTS)
    params_let_oil = slicedict(params, GO_LET_OIL + GO_OIL_ENDPOINTS)
    if set(GO_COREY_OIL).issubset(set(params_corey_oil)):
        gasoil.add_corey_oil(**params_corey_oil)
        logger.debug(
            "Added Corey gas to GasOil object from parameters %s",
            str(params_corey_oil.keys()),
        )
    elif set(GO_LET_OIL).issubset(set(params_let_oil)):
        params_let_oil["l"] = params_let_oil.pop("log")
        params_let_oil["e"] = params_let_oil.pop("eog")
        params_let_oil["t"] = params_let_oil.pop("tog")
        gasoil.add_LET_oil(**params_let_oil)
        logger.debug(
            "Added LET gas to GasOil object from parameters %s",
            str(params_let_oil.keys()),
        )
    else:
        logger.warning("Missing or ambiguous parameters for oil curve in GasOil object")
    if not gasoil.selfcheck():
        raise ValueError(
            ("Incomplete GasOil object, some parameters missing to factory")
        )

    return gasoil


def create_water_oil_gas(
    params: Optional[Dict[str, float]] = None, fast: bool = False
) -> WaterOilGas:
    """Create a WaterOilGas object from a dictionary of parameters

    Parameterization (Corey/LET) is inferred from presence
    of certain parameters in the dictionary.

    Check create_water_oil() and create_gas_oil() for lists
    of supported parameters (case insensitive)

    Params:
        params: Dictionary with parameters describing the WaterOilGas object.
        fast: If fast-mode should be set for constructed object.
    """
    if not params:
        params = {}
    if not isinstance(params, dict):
        raise TypeError("Parameter to create_water_oil_gas must be a dictionary")

    check_deprecated(params)

    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    wateroil: Optional[WaterOil]
    if sufficient_water_oil_params(params, failhard=False):
        wateroil = create_water_oil(params, fast=fast)
    else:
        logger.info("No wateroil parameters. Assuming only gas-oil in wateroilgas")
        wateroil = None

    # If the swl in WaterOil was initialized with swlheight,
    # ensure that result is passed on to the GasOil object:
    if "swl" not in params and "swlheight" in params and wateroil is not None:
        params["swl"] = wateroil.swl

    gasoil: Optional[GasOil]
    if sufficient_gas_oil_params(params, failhard=False):
        gasoil = create_gas_oil(params, fast=fast)
    else:
        logger.info("No gasoil parameters, assuming two-phase oilwatergas")
        gasoil = None

    wog_init_params = slicedict(params, WOG_INIT)
    wateroilgas = WaterOilGas(**wog_init_params, fast=fast)
    # The wateroilgas __init__ has already created WaterOil and GasOil objects
    # but we overwrite the references with newly created ones, this factory function
    # must then guarantee that they are compatible.
    wateroilgas.wateroil = wateroil  # This might be None
    wateroilgas.gasoil = gasoil  # This might be None
    if not wateroilgas.selfcheck():
        raise ValueError(f"Inconsistent WaterOilGas object. Bug? Input was {params}")
    return wateroilgas


def create_gas_water(
    params: Optional[Dict[str, float]] = None, fast: bool = False
) -> GasWater:
    """Create a GasWater object.

    Parameterization (Corey/LET) is inferred from presence
    of certain parameters in the dictionary.

    Args:
        params: Dictionary with parameters for GasWater.
        fast: If fast-mode should be set for constructed object.
    """
    if not params:
        params = {}
    if not isinstance(params, dict):
        raise TypeError("Parameter to create_gas_water must be a dictionary")

    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    sufficient_gas_water_params(params, failhard=True)

    gw_init_params = slicedict(params, GW_INIT)
    gaswater = GasWater(**gw_init_params, fast=fast)

    # We are using the create_water_oil_gas factory function
    # to avoid replicating code. It works because GasWater and
    # WaterOilGas internally are very similar.
    wog_params = params.copy()
    if "sgrw" in params:
        wog_params["sorw"] = params["sgrw"]
    # Set some dummy parameters for oil:
    wog_params["nog"] = 1
    wog_params["now"] = 1
    wog = create_water_oil_gas(wog_params, fast=fast)
    assert wog.wateroil is not None
    assert wog.gasoil is not None
    gaswater.wateroil = wog.wateroil
    gaswater.gasoil = wog.gasoil
    return gaswater


def create_scal_recommendation(
    params: Dict[str, Dict[str, float]],
    tag: str = "",
    h: Optional[float] = None,
    fast: bool = False,
) -> SCALrecommendation:
    """
    Set up a SCAL recommendation curve set from input as a
    dictionary of dictionary.

    The keys in in the dictionary *must* be "low", "base" and "high".

    The value for "low" must be a new dictionary with saturation
    endpoints and LET/Corey parameters, as you would feed it to
    the create_water_oil_gas() factory function, and then similarly
    for base and high.

    For oil-water only, you may omit the parameters for gas-oil.
    A WaterOilGas object for each case is created, but only the
    WaterOil part of it will be used.

    For gas-water, a GasWater object is created for each pess, base
    and high.

    Args:
        params: keys low, base and high.
            The value for "low" must be a new dictionary with saturation
            endpoints and LET/Corey parameters, as you would feed it to
            the create_water_oil_gas() factory function, and then similarly
            for base and high.
        tag: String to be used as the tag, will end up in comments.
        h: Saturation step length
        fast: If fast-mode should be set for constructed object.
    """
    if not isinstance(params, dict):
        raise ValueError("Input must be a dictionary (of dictionaries)")

    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    if "low" not in params:
        raise ValueError('"low" case not supplied')
    if "base" not in params:
        raise ValueError('"base" case not supplied')
    if "high" not in params:
        raise ValueError('"high" case not supplied')
    if not (
        isinstance(params["low"], dict)
        and isinstance(params["base"], dict)
        and isinstance(params["high"], dict)
    ):
        raise ValueError("All values in parameter dict must be dictionaries")

    check_deprecated(params["low"])
    check_deprecated(params["base"])
    check_deprecated(params["high"])

    errored = False
    if h is not None:
        params["low"]["h"] = h
        params["base"]["h"] = h
        params["high"]["h"] = h

    # Check parameter availability, in order to determine phase configuration
    gaswater = all(sufficient_gas_water_params(params[case]) for case in params)
    gasoil = all(sufficient_gas_oil_params(params[case]) for case in params)
    wateroil = all(sufficient_water_oil_params(params[case]) for case in params)

    wog_low: Union[WaterOilGas, GasWater]
    wog_base: Union[WaterOilGas, GasWater]
    wog_high: Union[WaterOilGas, GasWater]

    if wateroil or gasoil:
        try:
            wog_low = create_water_oil_gas(params["low"], fast=fast)
        except ValueError as err:
            raise ValueError(f"Problem with low/pess case: {err}") from err
        try:
            wog_base = create_water_oil_gas(params["base"], fast=fast)
        except ValueError as err:
            raise ValueError(f"Problem with base case: {err}") from err
        try:
            wog_high = create_water_oil_gas(params["high"], fast=fast)
        except ValueError as err:
            raise ValueError(f"Problem with high/opt case: {err}") from err
    elif gaswater:
        # Note that gaswater will be True in three-phase configs.
        try:
            wog_low = create_gas_water(params["low"], fast=fast)
        except ValueError as err:
            raise ValueError(f"Problem with low/pess case: {err}") from err
        try:
            wog_base = create_gas_water(params["base"], fast=fast)
        except ValueError as err:
            raise ValueError(f"Problem with base case: {err}") from err
        try:
            wog_high = create_gas_water(params["high"], fast=fast)
        except ValueError as err:
            raise ValueError(f"Problem with high/opt case: {err}") from err

    errored = all(not wog.selfcheck() for wog in [wog_low, wog_base, wog_high])

    if errored:
        raise ValueError("Incomplete SCAL recommendation")

    return SCALrecommendation(wog_low, wog_base, wog_high, tag)


def load_relperm_df(
    inputfile: Union[str, pd.DataFrame], sheet_name: Optional[str] = None
) -> pd.DataFrame:
    """Read CSV or XLSX from file and return scal/relperm data
    a dataframe.

    Checks validity in SATNUM and CASE columns.
    Ensures case-insensitiveness SATNUM, CASE, TAG and COMMENT

    Merges COMMENT into TAG column, as only TAG is picked up downstream.
    Adds a prefix "SATNUM <number>" to all tags.

    All strings in CASE column are converted to lowercase. Applies
    aliasing in the CASE column so that "pessimistic" and "pess" map to
    "low", and "optimistic" and "opt" map to "high".

    Args:
        inputfile: Filename for XLSX or CSV file, or a
            pandas DataFrame.
        sheet_name: Sheet-name, only used when loading xlsx files.

    Returns:
        To be handed over to pyscal list factory methods.
        Empty dataframe in case of errors (messages will be logged).
    """
    if isinstance(inputfile, (str, Path)) and Path(inputfile).is_file():
        tabular_file_format = infer_tabular_file_format(inputfile)
        if not tabular_file_format:
            raise ValueError(
                f"Impossible to infer file format for {inputfile}, not CSV/XLS/XLSX"
            )

        if tabular_file_format == "csv" and sheet_name is not None:
            logger.warning(
                "Sheet name only relevant for XLSX files, ignoring %s", sheet_name
            )
        excel_engines = {"xls": "xlrd", "xlsx": "openpyxl"}
        if tabular_file_format != "csv" and sheet_name:
            try:
                input_df = pd.read_excel(
                    inputfile,
                    sheet_name=sheet_name,
                    engine=excel_engines[tabular_file_format],
                )
                logger.info(
                    "Parsed %s file %s, sheet %s",
                    tabular_file_format.upper(),
                    inputfile,
                    sheet_name,
                )
            except (KeyError, ValueError) as err:
                raise ValueError(
                    f"Non-existing sheet-name {sheet_name} provided."
                ) from err
        elif tabular_file_format.startswith("xls"):
            input_df = pd.read_excel(
                inputfile, engine=excel_engines[tabular_file_format]
            )
            logger.info("Parsed %s file %s", tabular_file_format.upper(), inputfile)
        else:
            with open(inputfile, "r", encoding="utf-8") as csvfile:
                try:
                    delimiter = (
                        csv.Sniffer()
                        .sniff(csvfile.read(), [",", ";", "\t"])  # type: ignore
                        .delimiter
                    )
                except Exception:
                    # Sniffer cannot identify the delimiter
                    # typically on single column csv
                    delimiter = ","
                if delimiter != ",":
                    raise TypeError(
                        "Supplied file is not a valid CSV file. "
                        f"Detected delimiter is '{delimiter}'"
                    )
                csvfile.seek(0)
                input_df = pd.read_csv(csvfile, skipinitialspace=True, encoding="utf-8")
            logger.info("Parsed CSV file %s", inputfile)

    elif isinstance(inputfile, pd.DataFrame):
        input_df = inputfile
    else:
        if isinstance(inputfile, str) and not Path(inputfile).is_file():
            raise IOError("File not found " + str(inputfile))
        raise ValueError("Unsupported argument " + str(inputfile))
    assert isinstance(input_df, pd.DataFrame)

    if input_df.empty:
        logger.error("Relperm input dataframe is empty!")

    # We will ignore case on the column names, solved by converting column
    # name to uppercase
    ignorecasecolumns = ["satnum", "case", "tag", "comment"]
    for colname in input_df.columns:
        if colname.lower() in ignorecasecolumns:
            input_df = input_df.rename({colname: colname.upper()}, axis="columns")

    # Support both TAG and COLUMN (as TAG)
    if "COMMENT" in input_df and "TAG" not in input_df:
        input_df = input_df.rename({"COMMENT": "TAG"}, axis="columns")
    if "COMMENT" in input_df and "TAG" in input_df:
        # It might never happen that a user puts both tag and comment in
        # the dataframe, but if so, merge them.
        input_df["TAG"] = (
            "tag: " + input_df["TAG"] + "; comment: " + input_df["COMMENT"]
        )
    # It is tempting to detect any NaN's at this point because that can
    # indicate merged cells, which is not supported, but that could
    # break optional comment columns which might only be defined on certain lines.
    if "SATNUM" not in input_df:
        raise ValueError("SATNUM must be present in CSV/XLSX file/dataframe")

    # Delete columns and rows that are all NaNs (this has been observed
    # to occur from seemingly empty cells in Excel/LibreOffice and
    # has been seen to vary with Pandas/Openpyxl versions in use)
    input_df = input_df.dropna(axis="columns", how="all")
    input_df = input_df.dropna(axis="index", how="all")

    if input_df["SATNUM"].isna().sum() > 0:
        raise ValueError(
            "Found not-a-number in the SATNUM column. This could be due to "
            "merged cells in XLSX, which is not supported."
        )

    # Check that SATNUM's are consecutive and integers:
    try:
        input_df["SATNUM"] = input_df["SATNUM"].astype(int)
    except ValueError as err:
        raise ValueError("SATNUM must contain only integers") from err

    if min(input_df["SATNUM"]) != 1:
        raise ValueError("SATNUM must start at 1")

    if max(input_df["SATNUM"]) != len(input_df["SATNUM"].unique()):
        raise ValueError(
            "Missing SATNUMs? Max SATNUM is not equal to number of unique SATNUMS"
        )
    if "CASE" not in input_df and len(input_df["SATNUM"].unique()) != len(input_df):
        raise ValueError("Non-unique SATNUMs?")
    # If we are in a SCAL recommendation setting
    if "CASE" in input_df:
        # Enforce lower case:
        if input_df["CASE"].isna().sum() > 0:
            raise ValueError(
                "Found not-a-number in the CASE column. This could be due "
                "merged cells in XLSX, which is not supported."
            )
        input_df["CASE"] = remap_validate_cases(input_df["CASE"].to_numpy())

    # Add the SATNUM index to the TAG column
    if "TAG" not in input_df:
        input_df["TAG"] = ""
    input_df["TAG"] = input_df["TAG"].fillna(value="")
    input_df["TAG"] = "SATNUM " + input_df["SATNUM"].astype(str) + " " + input_df["TAG"]

    # Check if fast is defined as a column
    if "fast" in input_df:
        logger.warning("Fast mode is not an option for individual SATNUMs")
        logger.warning("it is implemented as a global option.")
        logger.warning("The fast column in the dataframe will be ignored.")
        logger.warning("Use fast=True in the function call instead.")
        logger.warning("Fast mode is only available through the Python API")

    # Check that we are able to make something out of the first row:
    firstrow = input_df.iloc[0, :]
    error: bool = False
    wo_ok = sufficient_water_oil_params(firstrow)
    go_ok = sufficient_gas_oil_params(firstrow)
    gw_ok = sufficient_gas_water_params(firstrow)
    if error or (not wo_ok and not go_ok and not gw_ok):
        raise ValueError(
            "Can't make neither WaterOil, GasOil or GasWater from "
            "the given data. Check documentation for what you need to supply. "
            f"You provided the columns {input_df.columns.to_numpy()}"
        )
    logger.info(
        "Loaded input data with %s SATNUMS, column %s",
        str(len(input_df["SATNUM"].unique())),
        str(input_df.columns.to_numpy()),
    )
    return input_df.sort_values("SATNUM")


def alias_sgrw(params: Dict[str, Any]) -> Dict[str, Any]:
    """Allow sgrw as an alias for sorw by remapping a
    sgrw value to a sorw value in an incoming dict.

    Will error hard of sorw already exists and is not nan.

    This aliasing is relevant when GasWater is modelled
    as three-phase to allow for condensate forming, and mirrors
    GasWater.__init__() which performs the same aliasing.

    Args:
        params: Keys must be lower case.
    """
    if "sgrw" in params:
        if "sorw" not in params or pd.isna(params["sorw"]):
            params_copy = dict(params)
            params_copy["sorw"] = params["sgrw"]
            del params_copy["sgrw"]
            return params_copy
        if np.isclose(params["sgrw"], params["sorw"]):
            params_copy = dict(params)
            del params_copy["sgrw"]
            return params_copy
        raise ValueError(
            f"sgrw ({params['sgrw']}) must equal sorw ({params['sorw']}) "
            "when both are supplied to WaterOil."
        )
    return params


def remap_validate_cases(casevalues: List[str]) -> List[str]:
    """Remap values in the CASE column so that we can use aliases.

    All values are first made lower case, then
    "pessimistic" and "pess" are mapped to "low" and
    "optimistic" and "opt" are mapped to "high".

    Will raise ValueError if some values are not understood, and
    if we don't have exactly three unique values.

    Args:
        casevalues: values to remap.
    """
    accepted = ["low", "base", "high"]
    aliases = {
        "pessimistic": "low",
        "pess": "low",
        "optimistic": "high",
        "opt": "high",
    }
    lowered = [value.lower() for value in casevalues]
    remapped = [aliases.get(value, value) for value in lowered]
    not_understood = set(remapped) - set(accepted)
    if not_understood:
        raise ValueError(f"Invalid case values: {not_understood}")
    if len(set(remapped)) != len(accepted):
        raise ValueError(
            f"You must supply low, base AND high, got only {set(remapped)}"
        )
    return remapped


def create_scal_recommendation_list(
    input_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
) -> PyscalList:
    """Requires SATNUM and CASE to be defined in the input data

    Args:
        input_df: Input data, should have been processed
            through load_relperm_df().
        h: Saturation step-value
        fast: If fast-mode should be set for constructed object

    Returns:
        PyscalList, consisting of SCALrecommendation objects
    """
    scal_l = PyscalList()
    assert isinstance(input_df, pd.DataFrame)

    scalinput = input_df.set_index(["SATNUM", "CASE"])

    for satnum in scalinput.index.levels[0].to_numpy():
        # load_relperm_df only validates the CASE column for all SATNUMs at
        # once, errors for particular SATNUMs are caught here.
        if len(scalinput.loc[satnum, :]) > 3:
            raise ValueError(f"Too many cases supplied for SATNUM {satnum}")
        if len(scalinput.loc[satnum, :]) < 3:
            raise ValueError(f"Too few cases supplied for SATNUM {satnum}")
        try:
            scal_l.append(
                create_scal_recommendation(
                    scalinput.loc[satnum, :].to_dict(orient="index"), h=h, fast=fast
                )
            )
        except ValueError as err:
            raise ValueError(f"Error for SATNUM {satnum}: {err}") from err

    return scal_l


def create_pyscal_list(
    relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
):
    """Create WaterOilGas, WaterOil, GasOil or GasWater list
    based on what is available

    Args:
        relperm_params_df: Input data, should have been processed
            through load_relperm_df().
        h: Saturation step-value
        fast: If fast-mode should be set for constructed object

    Returns:
        PyscalList, consisting of either WaterOil, GasOil or WaterOilGas objects
    """
    params = relperm_params_df.iloc[0, :]  # first row
    water_oil = sufficient_water_oil_params(params)
    gas_oil = sufficient_gas_oil_params(params)
    gas_water = sufficient_gas_water_params(params)

    if water_oil and gas_oil:
        return create_wateroilgas_list(relperm_params_df, h, fast)
    if water_oil:
        return create_wateroil_list(relperm_params_df, h, fast)
    if gas_oil:
        return create_gasoil_list(relperm_params_df, h, fast)
    if gas_water:
        return create_gaswater_list(relperm_params_df, h, fast)
    raise ValueError("Could not determine two or three phase from parameters")


def create_wateroilgas_list(
    relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
) -> PyscalList:
    """Create a PyscalList with WaterOilGas objects from
    a dataframe

    Args:
        relperm_params_df: Input data, should have been processed
            through load_relperm_df().
        h: Saturation step-value
        fast: If fast-mode should be set for constructed object

    Returns:
        PyscalList, consisting of WaterOilGas objects
    """
    wogl = PyscalList()
    for row_idx, params in relperm_params_df.sort_values("SATNUM").iterrows():
        if h is not None:
            params["h"] = h
        try:
            wogl.append(create_water_oil_gas(params.to_dict(), fast=fast))
        except (AssertionError, ValueError, TypeError) as err:
            raise ValueError(f"Error for SATNUM {row_idx + 1}: {err}") from err
    return wogl


def create_wateroil_list(
    relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
) -> PyscalList:
    """Create a PyscalList with WaterOil objects from
    a dataframe

    Args:
        relperm_params_df: A valid dataframe with
            WaterOil parameters, processed through load_relperm_df()
        h: Saturation steplength
        fast: If fast-mode should be set for constructed object

    Returns:
        PyscalList, consisting of WaterOil objects
    """
    wol = PyscalList()
    for _, params in relperm_params_df.iterrows():
        if h is not None:
            params["h"] = h
        try:
            wol.append(create_water_oil(params.to_dict(), fast=fast))
        except (AssertionError, ValueError, TypeError) as err:
            raise ValueError(f"Error for SATNUM {params['SATNUM']}: {err}") from err
    return wol


def create_gasoil_list(
    relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
) -> PyscalList:
    """Create a PyscalList with GasOil objects from
    a dataframe

    Args:
        relperm_params_df: A valid dataframe with GasOil parameters,
            processed through load_relperm_df()
        h: Saturation steplength
        fast: If fast-mode should be set for constructed object

    Returns:
        PyscalList, consisting of GasOil objects
    """
    gol = PyscalList()
    for _, params in relperm_params_df.iterrows():
        if h is not None:
            params["h"] = h
        try:
            gol.append(create_gas_oil(params.to_dict(), fast=fast))
        except (AssertionError, ValueError, TypeError) as err:
            raise ValueError(f"Error for SATNUM {params['SATNUM']}: {err}") from err
    return gol


def create_gaswater_list(
    relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
) -> PyscalList:
    """Create a PyscalList with WaterOilGas objects from
    a dataframe, to be used for GasWater

    Args:
        relperm_params_df: A valid dataframe with GasWater
            parameters, processed through load_relperm_df()
        h: Saturation steplength
        fast: If fast-mode should be set for constructed object

    Returns:
        PyscalList, consisting of GasWater objects
    """
    gwl = PyscalList()
    for _, params in relperm_params_df.iterrows():
        if h is not None:
            params["h"] = h
        try:
            gwl.append(create_gas_water(params.to_dict(), fast=fast))
        except (AssertionError, ValueError, TypeError) as err:
            raise ValueError(f"Error for SATNUM {params['SATNUM']}: {err}") from err
    return gwl


class PyscalFactory:
    """Class for implementing the factory pattern for Pyscal objects

    The factory functions herein can take multiple parameter sets,
    determine what kind of parametrization to be used, and set up
    the full objects based on these parameters, instead of
    explicitly having to call the API for each task.

    Example::

      wo = WaterOil(sorw=0.05)
      wo.add_corey_water(nw=3)
      wo.add_corey_oil(now=2)
      # is equivalent to:
      wo = factory.create_water_oil(dict(sorw=0.05, nw=3, now=2))

    Parameter names to factory functions are case *insensitive*, while
    the add_*() parameters are not. This is because the add_*() parameters
    are meant as a Python API, while the factory class is there to aid
    users when input is written in a different context, like an Excel
    spreadsheet.
    """

    @staticmethod
    def create_water_oil(
        params: Optional[Dict[str, float]] = None, fast: bool = False
    ) -> WaterOil:
        warnings.warn(
            "PyscalFactory.create_water_oil is deprecated. "
            "Please use create_water_oil function instead.",
            DeprecationWarning,
        )
        return create_water_oil(params, fast)

    @staticmethod
    def create_gas_oil(
        params: Optional[Dict[str, float]] = None, fast: bool = False
    ) -> GasOil:
        warnings.warn(
            "PyscalFactory.create_gas_oil is deprecated. "
            "Please use create_gas_oil function instead.",
            DeprecationWarning,
        )
        return create_gas_oil(params, fast)

    @staticmethod
    def create_water_oil_gas(
        params: Optional[Dict[str, float]] = None, fast: bool = False
    ) -> WaterOilGas:
        warnings.warn(
            "PyscalFactory.create_water_oil_gas is deprecated. "
            "Please use create_water_oil_gas function instead.",
            DeprecationWarning,
        )
        return create_water_oil_gas(params, fast)

    @staticmethod
    def create_gas_water(
        params: Optional[Dict[str, float]] = None, fast: bool = False
    ) -> GasWater:
        warnings.warn(
            "PyscalFactory.create_gas_water is deprecated. "
            "Please use create_gas_water function instead.",
            DeprecationWarning,
        )
        return create_gas_water(params, fast)

    @staticmethod
    def create_scal_recommendation(
        params: Dict[str, Dict[str, float]],
        tag: str = "",
        h: Optional[float] = None,
        fast: bool = False,
    ) -> SCALrecommendation:
        warnings.warn(
            "PyscalFactory.create_scal_recommendation is deprecated. "
            "Please use create_scal_recommendation instead.",
            DeprecationWarning,
        )
        return create_scal_recommendation(params, tag, h, fast)

    @staticmethod
    def load_relperm_df(
        inputfile: Union[str, pd.DataFrame], sheet_name: Optional[str] = None
    ) -> pd.DataFrame:
        warnings.warn(
            "PyscalFactory.load_relperm_df is deprecated. "
            "Please use load_relperm_df instead.",
            DeprecationWarning,
        )
        return load_relperm_df(inputfile, sheet_name)

    @staticmethod
    def alias_sgrw(params: Dict[str, Any]) -> Dict[str, Any]:
        warnings.warn(
            "PyscalFactory.alias_sgrw is deprecated. Please use alias_sgrw instead.",
            DeprecationWarning,
        )
        return alias_sgrw(params)

    @staticmethod
    def remap_validate_cases(casevalues: List[str]) -> List[str]:
        warnings.warn(
            "PyscalFactory.remap_validate_cases is deprecated. "
            "Please use remap_validate_cases instead.",
            DeprecationWarning,
        )
        return remap_validate_cases(casevalues)

    @staticmethod
    def create_scal_recommendation_list(
        input_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
    ) -> PyscalList:
        warnings.warn(
            "PyscalFactory.create_scal_recommendation_list is deprecated. "
            "Please use create_scal_recommendation_list instead.",
            DeprecationWarning,
        )
        return create_scal_recommendation_list(input_df, h, fast)

    @staticmethod
    def create_pyscal_list(
        relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
    ):
        warnings.warn(
            "PyscalFactory.create_pyscal_list is deprecated. "
            "Please use create_pyscal_list instead.",
            DeprecationWarning,
        )
        return create_pyscal_list(relperm_params_df, h, fast)

    @staticmethod
    def create_wateroilgas_list(
        relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
    ) -> PyscalList:
        warnings.warn(
            "PyscalFactory.create_wateroilgas_list is deprecated. "
            "Please use create_wateroilgas_list instead.",
            DeprecationWarning,
        )
        return create_wateroilgas_list(relperm_params_df, h, fast)

    @staticmethod
    def create_wateroil_list(
        relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
    ) -> PyscalList:
        warnings.warn(
            "PyscalFactory.create_wateroil_list is deprecated. "
            "Please use create_wateroil_list instead.",
            DeprecationWarning,
        )
        return create_wateroil_list(relperm_params_df, h, fast)

    @staticmethod
    def create_gasoil_list(
        relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
    ) -> PyscalList:
        warnings.warn(
            "PyscalFactory.create_gasoil_list is deprecated. "
            "Please use create_gasoil_list instead.",
            DeprecationWarning,
        )
        return create_gasoil_list(relperm_params_df, h, fast)

    @staticmethod
    def create_gaswater_list(
        relperm_params_df: pd.DataFrame, h: Optional[float] = None, fast: bool = False
    ) -> PyscalList:
        warnings.warn(
            "PyscalFactory.create_gaswater_list is deprecated. "
            "Please use create_gaswater_list instead.",
            DeprecationWarning,
        )
        return create_gaswater_list(relperm_params_df, h, fast)


def sufficient_water_oil_params(params: dict, failhard: bool = False) -> bool:
    """Determine if the supplied parameters are sufficient for
    attempting creating a WaterOil object.

    In the factory context, relying on the defaults in the API
    is not allowed, as that would leave the tasks for the factory
    undefined (Corey or LET, and which pc?)

    Args:
        params: Dictionary of parameters to a WaterOil object.
        failhard: If True, will raise ValueError when
            parameters are insufficient. If defaulted, no
            exception is raised.

    Returns:
        True if a WaterOil object should be attempted constructed
        (but no guarantee for validity of numerical values)
    """
    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    water_ok = (
        len(slicedict(params, set(WO_COREY_WATER))) == 1
        or len(slicedict(params, set(WO_LET_WATER))) == 3
    )
    oil_ok = (
        len(slicedict(params, set(WO_COREY_OIL))) == 1
        or len(slicedict(params, set(WO_LET_OIL))) == 3
        or len(slicedict(params, set(WO_LET_OIL_ALT))) == 3
    )
    if failhard and not (water_ok and oil_ok):
        raise ValueError("Missing WaterOil parameters in " + str(params))
    return water_ok and oil_ok


def sufficient_gas_oil_params(params: dict, failhard: bool = False) -> bool:
    """Determine if the supplied parameters are sufficient for
    attempting at creating a GasOil object.

    In the factory context, relying on the defaults in the API
    is not allowed, as that would leave the tasks for the factory
    undefined (Corey or LET, and which pc?)

    Args:
        params: Dictionary of parameters to a GasOil object.
        failhard: If True, will raise ValueError when
            parameters are insufficient. If defaulted, no
            exception is raised.

    Returns:
        True if a GasOil object should be attempted constructed
        (but no guarantee for validity of numerical values)
    """
    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    oil_ok = (
        len(slicedict(params, set(GO_COREY_OIL))) == 1
        or len(slicedict(params, set(GO_LET_OIL))) == 3
    )
    gas_ok = (
        len(slicedict(params, set(GO_COREY_GAS))) == 1
        or len(slicedict(params, set(GO_LET_GAS))) == 3
    )
    if failhard and not (oil_ok and gas_ok):
        raise ValueError("Missing gas-oil parameters in " + str(params))
    return oil_ok and gas_ok


def sufficient_gas_water_params(params: dict, failhard: bool = False) -> bool:
    """Determine if the supplied parameters are sufficient for
    attempting creating a WaterOilGas object to be used for gas water.

    In the factory context, relying on the defaults in the API
    (wateroilgas.py) is not allowed, as that would leave the tasks
    for the factory undefined (Corey or LET, and which pc?)

    Args:
        params: Dictionary of parameters to a GasWater object.
        failhard: If True, will raise ValueError when
            parameters are insufficient. If defaulted, no
            exception is raised.

    Returns:
        True if a GasWater object should be attempted constructed
            (but no guarantee for validity of numerical values)
    """
    # For case insensitiveness, all keys are converted to lower case:
    params = {key.lower(): value for (key, value) in params.items()}

    water_ok = (
        len(slicedict(params, set(GW_COREY_WATER))) == 1
        or len(slicedict(params, set(GW_LET_WATER))) == 3
    )
    gas_ok = (
        len(slicedict(params, set(GW_COREY_GAS))) == 1
        or len(slicedict(params, set(GW_LET_GAS))) == 3
    )
    if failhard and not (water_ok and gas_ok):
        raise ValueError("Missing gas-water parameters in " + str(params))
    return water_ok and gas_ok


def filter_nan_from_dict(params: dict) -> dict:
    """Clean out keys with NaN values in a dict.

    Key with string values are passed through (empty strings are allowed)

    Args:
        params (dict): Any dictionary

    Returns
        dict, with as many or fewer keys.
    """
    cleaned_params = {}
    for key, value in params.items():
        if isinstance(value, str) or not np.isnan(value):
            cleaned_params[key] = value
    return cleaned_params


def infer_tabular_file_format(filename: Union[str, Path]) -> str:
    """Determine the file format of a file containing tabular data,
    distinguishes between csv, xls and xlsx

    Args:
        filename: Path to file

    Returns:
        One of "csv", "xlsx" or "xls". Empty string if nothing found out.
    """
    try:
        pd.read_excel(filename, engine="openpyxl")
        return "xlsx"
    except (
        ValueError,
        OSError,
        openpyxl.utils.exceptions.InvalidFileException,
        zipfile.BadZipFile,
    ):
        # < Pandas 1.2, we get InvalidFileException from openpyxl
        # Pandas 1.2.0 - 1.2.1: CSV gives ValueError,
        #                       XLS gives OSError
        # >= Pandas 1.2.2:      CSV gives zipfile.BadZipFile,
        #                       XLS gives OSError
        pass
    try:
        pd.read_excel(filename, engine="xlrd")
        return "xls"
    except (ValueError, TypeError, xlrd.biffh.XLRDError):
        # We get here for both CSV and XLSX files.
        pass
    try:
        dframe = pd.read_csv(filename, encoding="utf-8")
        # >= 1.2.1: Pandas: Bugfix makes read_csv() more encoding fault
        #                   tolerant which this code opts out from.
        if not dframe.empty:
            return "csv"
    except UnicodeDecodeError:
        # (xls and xlsx files)
        pass
    except (pd.errors.ParserError, pd.errors.EmptyDataError) as csverror:
        logger.error("Message from CSV parser: %s", str(csverror))
        # (Some text file that is not CSV)

    return ""


def check_deprecated(params: Dict[str, Any]) -> None:
    """Check for deprecated parameter names

    Args:
        params: Dictionary of parameters for which only the keys are used here.
    """
    # Block long deprecated parameters with an exception.
    # Remove this block in pyscal 1.x
    if "krowend" in params and "kroend" not in params:
        raise ValueError("krowend is not supported by pyscal. Use kroend")
    if "krogend" in params and "kroend" not in params:
        raise ValueError("krogend is not supported by pyscal. Use kroend")



"""Utility function for pyscal"""

import logging
from typing import Callable, Optional, Tuple, Union

import numpy as np
from scipy.interpolate import interp1d

from pyscal import GasOil, WaterOil

logger = logging.getLogger(__name__)


def normalize_nonlinpart_wo(curve: WaterOil) -> Tuple[Callable, Callable]:
    """Make krw and krow functions that evaluate only on the
    (potentially) nonlinear part of the relperm curves, and with
    a normalized argument (0,1) on that interval.

    For a WaterOil krw curve, the nonlinear part is from swcr to sorw.
    swcr is mapped to zero, and 1 - sorw is mapped to 1. Then there is
    an assumed linear part from sorw to 1 which we ignore here.

    For a WaterOil krow curve, the nonlinear part
    is from 1 - sorw (mapped to zero) to swl (mapped to 1).

    These endpoints must be known the the WaterOil object coming in (the object
    can determine them using functions estimate_sorw(), estimate_swcr() and
    estimate_socr()

    If the entire curve is linear, it will not matter for this function, because
    this function only deals with the presumably known endpoints.

    Arguments:
        curve: incoming oilwater curve set (krw and krow)

    Returns:
        tuple of lambda functions. The first will evaluate krw on
        the normalized Sw interval [0,1], the second will
        evaluate krow on the normalized So interval [0,1].
    """
    krw_interp = interp1d(
        curve.table["SW"],
        curve.table["KRW"],
        kind="linear",
        bounds_error=False,
        fill_value=(0.0, curve.table["KRW"].max()),
    )

    # The internal dataframe might contain normalized
    # saturation values, but we do not want to assume they
    # are there or even correct, therefore we effectively
    # recalculate them
    def sw_fn(swn):
        return curve.swcr + swn * (1.0 - curve.swcr - curve.sorw)

    def krw_fn(swn):
        return krw_interp(sw_fn(swn))

    kro_interp = interp1d(
        1.0 - curve.table["SW"],
        curve.table["KROW"],
        kind="linear",
        bounds_error=False,
        fill_value=(0.0, curve.table["KROW"].max()),
    )

    def so_fn(son):
        return curve.socr + son * (1.0 - curve.socr - curve.swl)

    def kro_fn(son):
        return kro_interp(so_fn(son))

    return (krw_fn, kro_fn)


def normalize_nonlinpart_go(curve: GasOil) -> Tuple[Callable, Callable]:
    """Make krg and krog functions that evaluates only on the
    (potentially) nonlinear part of the relperm curves, and with
    a normalized argument (0,1) on that interval.

    For a GasOil krg curve, the nonlinear part
    is from sgcr to sorg. sgcr is mapped to sg=zero, and sg=1 - sorg - swl is mapped
    to 1. Then there is an assumed linear part from sorg to 1 which we ignore here.

    For a GasOil krog curve, the nonlinear part
    is from 1 - sorg (mapped to zero) to sg=sgro (mapped to 1).

    These endpoints must be known the the GasOil object coming in (the object
    can determine them using functions 'estimate_sorg()', 'estimate_sgcr()' and
    'estimate_sgro()'

    If the entire curve is linear, it will not matter for this function, because
    this function only deals with the presumably known endpoints.

    Arguments:
        curve: incoming gasoil curve set (krg and krog)

    Returns:
        tuple of functions. The first will evaluate krg on
        the normalized Sg interval [0,1], the second will
        evaluate krog on the normalized So interval [0,1].
    """
    krg_interp = interp1d(
        curve.table["SG"],
        curve.table["KRG"],
        kind="linear",
        bounds_error=False,
        fill_value=(0.0, curve.table["KRG"].max()),
    )

    # The internal dataframe might contain normalized
    # saturation values, but we do not want to assume they
    # are there or even correct, therefore we effectively
    # recalculate them
    def sg_fn(sgn):
        return curve.sgcr + sgn * (1.0 - curve.swl - curve.sgcr - curve.sorg)

    def krg_fn(sgn):
        return krg_interp(sg_fn(sgn))

    kro_interp = interp1d(
        1.0 - curve.table["SG"],
        curve.table["KROG"],
        kind="linear",
        bounds_error=False,
        fill_value=(0.0, curve.table["KROG"].max()),
    )

    def so_fn(son):
        return (
            curve.swl + curve.sorg + son * (1.0 - curve.swl - curve.sorg - curve.sgro)
        )

    def kro_fn(son):
        return kro_interp(so_fn(son))

    return (krg_fn, kro_fn)


def normalize_pc(curve: Union[WaterOil, GasOil]) -> Callable:
    """Normalize the capillary pressure curve.

    This is only normalized with respect to the
    smallest and largest saturation present in the table,
    not to the could-be-uncertain swirr
    that the object could contain, because we then have
    to make assumptions on the equations used to generate
    the data in the table.

    Args:
        curve: An object with a table with a pc column

    Returns:
        a lambda function that will evaluate pc on
        the normalized interval [0,1]
    """
    if isinstance(curve, WaterOil):
        sat_col = "SW"
    elif isinstance(curve, GasOil):
        sat_col = "SG"
    else:
        raise ValueError("Only WaterOil or GasOil allowed as argument")

    if "PC" not in curve.table:
        # Return a dummy zero lambda
        return lambda sxn: 0

    min_pc = curve.table["PC"].min()
    max_pc = curve.table["PC"].max()
    min_sx = curve.table[sat_col].min()
    max_sx = curve.table[sat_col].max()

    pc_interp = interp1d(
        curve.table[sat_col],
        curve.table["PC"],
        kind="linear",
        bounds_error=False,
        fill_value=(max_pc, min_pc),  # This gives constant extrapolation outside [0, 1]
    )

    # Map from normalized value to real saturation domain:
    def sx_fn(sxn):
        return curve.table[sat_col].min() + sxn * (max_sx - min_sx)

    def pc_fn(sxn):
        return pc_interp(sx_fn(sxn))

    return pc_fn


def _interpolate_tags(
    low: Union[WaterOil, GasOil],
    high: Union[WaterOil, GasOil],
    parameter: float,
    tag: Optional[str],
):
    """Preserve tag/comment. Depending on context, the
    interpolation parameter may or may not make sense. In a SCALrecommendation
    interpolation, the new tag should be constructed in the caller of this function.
    because of the way the parameter value is handled.

    This function is used by interpolate_wo and interpolate_go

    Args:
        low: low case in interpolation
        high: high case
        parameter: between 0 and 1
        tag: If not none, this is directly returned.

    Returns:
        string, a "computed" tag if a tag is not directly supplied
    """
    if tag is None:
        if low.tag == high.tag:
            if low.tag:
                return f"Interpolated to {parameter} in {low.tag}"
            # No tags defined.
            return f"Interpolated to {parameter}"
        return f"Interpolated to {parameter} between {low.tag} and {high.tag}"
    return tag


def interpolate_wo(
    wo_low: WaterOil,
    wo_high: WaterOil,
    parameter: float,
    h: Optional[float] = None,
    tag: Optional[str] = None,
) -> WaterOil:
    """Interpolates between two water-oil curves.

    The saturation endpoints for the curves must be known
    by the objects. They can be estimated by estimate_sorw() etc.
    or can be set manually for finer control.

    The interpolation algorithm is different left and right
    for saturation endpoints, and saturation endpoints are
    interpolated individually.

    Arguments:
        wo_low: a "low" case
        wo_high: a "high" case
        parameter: Between 0 and 1. 0 will return the low case, 1 will return
            the high case. Any number in between will return an interpolated curve
        h: Saturation step-size in interpolant. If defaulted, a value
            smaller than in the input curves are used, to preserve information.
        tag: Tag to associate to the constructed object. If None
            it will be automatically filled. Set to empty string to ensure no tag.
    """
    # Warning: Almost code duplication with corresponding _go function

    assert isinstance(wo_low, WaterOil)
    assert isinstance(wo_high, WaterOil)

    assert 0 <= parameter <= 1
    # Extrapolation is refused, but perhaps later implemented with truncation to (0,1)

    # Running fast mode if both interpolants have fast mode
    fast = wo_low.fast and wo_high.fast

    # Constructs functions that works on normalized saturation interval
    krw1, kro1 = normalize_nonlinpart_wo(wo_low)
    krw2, kro2 = normalize_nonlinpart_wo(wo_high)
    pc1 = normalize_pc(wo_low)
    pc2 = normalize_pc(wo_high)

    # Construct a function that can be applied to both relperm values
    # and endpoints
    def weighted_value(a, b):
        return a * (1.0 - parameter) + b * parameter

    # Interpolate saturation endpoints
    swl_new = weighted_value(wo_low.swl, wo_high.swl)
    swcr_new = weighted_value(wo_low.swcr, wo_high.swcr)
    sorw_new = weighted_value(wo_low.sorw, wo_high.sorw)
    socr_new = weighted_value(wo_low.socr, wo_high.socr)

    # Interpolate kr at saturation endpoints
    krwmax_new = weighted_value(wo_low.table["KRW"].max(), wo_high.table["KRW"].max())
    krwend_new = weighted_value(krw1(1), krw2(1))
    kroend_new = weighted_value(kro1(1), kro2(1))

    # Construct the new WaterOil object, with interpolated
    # endpoints:
    wo_new = WaterOil(
        swl=swl_new, swcr=swcr_new, sorw=sorw_new, socr=socr_new, h=h, fast=fast
    )

    # Add interpolated relperm data in nonlinear parts:
    wo_new.table["KRW"] = weighted_value(
        krw1(wo_new.table["SWN"]), krw2(wo_new.table["SWN"])
    )
    wo_new.table["KROW"] = weighted_value(
        kro1(wo_new.table["SON"]), kro2(wo_new.table["SON"])
    )

    wo_new.set_endpoints_linearpart_krw(krwend=krwend_new, krwmax=krwmax_new)
    wo_new.set_endpoints_linearpart_krow(kroend=kroend_new)

    # We need a new fit-for-purpose normalized swnpc, that ignores
    # the initial swnpc (swirr-influenced)
    wo_new.table["swn_pc_intp"] = (wo_new.table["SW"] - wo_new.table["SW"].min()) / (
        wo_new.table["SW"].max() - wo_new.table["SW"].min()
    )
    wo_new.table["PC"] = weighted_value(
        pc1(wo_new.table["swn_pc_intp"]), pc2(wo_new.table["swn_pc_intp"])
    )

    wo_new.tag = _interpolate_tags(wo_low, wo_high, parameter, tag)

    return wo_new


def interpolate_go(
    go_low: GasOil,
    go_high: GasOil,
    parameter: float,
    h: Optional[float] = None,
    tag: Optional[str] = None,
) -> GasOil:
    """Interpolates between two gas-oil curves.

    The saturation endpoints for the curves must be known
    by the objects. They can be estimated by estimate_sorg() etc.
    or can be set manually for finer control.

    The interpolation algorithm is different left and right
    for saturation endpoints, and saturation endpoints are
    interpolated individually.

    Arguments:
        go_low: a "low" case
        go_high: a "high" case
        parameter: Between 0 and 1. 0 will return the low case, 1 will return
            the high case. Any number in between will return an interpolated curve
        h: Saturation step-size in interpolant. If defaulted, a value
            smaller than in the input curves are used, to preserve information.
        tag: Tag to associate to the constructed object. If None
            it will be automatically filled. Set to empty string to ensure no tag.
    """
    # Warning: Almost code duplication with corresponding _wo function

    assert isinstance(go_low, GasOil)
    assert isinstance(go_high, GasOil)

    assert 0 <= parameter <= 1
    # Extrapolation is refused, but perhaps later implemented with truncation to (0,1)

    # Running fast mode if both interpolants have fast mode
    fast = go_low.fast and go_high.fast

    # Constructs functions that works on normalized saturation interval
    krg1, kro1 = normalize_nonlinpart_go(go_low)
    krg2, kro2 = normalize_nonlinpart_go(go_high)
    pc1 = normalize_pc(go_low)
    pc2 = normalize_pc(go_high)

    # Construct a lambda function that can be applied to both relperm values
    # and endpoints
    def weighted_value(a, b):
        return a * (1.0 - parameter) + b * parameter

    # Interpolate saturation endpoints
    swl_new = weighted_value(go_low.swl, go_high.swl)
    sgcr_new = weighted_value(go_low.sgcr, go_high.sgcr)
    sorg_new = weighted_value(go_low.sorg, go_high.sorg)
    sgro_new = weighted_value(go_low.sgro, go_high.sgro)

    if not (np.isclose(sgro_new, sgcr_new) or np.isclose(sgro_new, 0.0)):
        raise ValueError(
            f"Interpolated sgro ({sgro_new}) not equal "
            f"to zero or interpolated sgcr ({sgcr_new})"
        )

    # Interpolate kr at saturation endpoints
    krgmax_new = weighted_value(go_low.table["KRG"].max(), go_high.table["KRG"].max())
    krgend_new = weighted_value(krg1(1), krg2(1))
    kromax_new = weighted_value(go_low.table["KROG"].max(), go_high.table["KROG"].max())
    kroend_new = weighted_value(kro1(1), kro2(1))

    # Construct the new GasOil object, with interpolated
    # endpoints:
    go_new = GasOil(
        swl=swl_new, sgcr=sgcr_new, sorg=sorg_new, sgro=sgro_new, h=h, fast=fast
    )

    # Add interpolated relperm data in nonlinear parts:
    go_new.table["KRG"] = weighted_value(
        krg1(go_new.table["SGN"]), krg2(go_new.table["SGN"])
    )
    go_new.table["KROG"] = weighted_value(
        kro1(go_new.table["SON"]), kro2(go_new.table["SON"])
    )
    go_new.table["PC"] = weighted_value(
        pc1(go_new.table["SGN"]), pc2(go_new.table["SGN"])
    )

    # We need a new fit-for-purpose normalized sgnpc
    go_new.table["sgn_pc_intp"] = (go_new.table["SG"] - go_new.table["SG"].min()) / (
        go_new.table["SG"].max() - go_new.table["SG"].min()
    )
    go_new.table["PC"] = weighted_value(
        pc1(go_new.table["sgn_pc_intp"]), pc2(go_new.table["sgn_pc_intp"])
    )

    go_new.set_endpoints_linearpart_krog(kroend=kroend_new, kromax=kromax_new)

    # Here we should have honored krgendanchor. Check github issue.
    go_new.set_endpoints_linearpart_krg(krgend=krgend_new, krgmax=krgmax_new)

    go_new.tag = _interpolate_tags(go_low, go_high, parameter, tag)

    return go_new



import subprocess
import sys
from pathlib import Path

import pytest

import pyscal
from pyscal import pyscalcli


def test_default_logger_levels_and_split(capsys):
    """Verify that the intended usage of this logger have expected results"""

    splitlogger = pyscal.getLogger_pyscal("test_levels_split")

    splitlogger.debug("This DEBUG-text is not to be seen")
    captured = capsys.readouterr()
    assert "DEBUG-text" not in captured.out
    assert "DEBUG-text" not in captured.err

    splitlogger.info("This INFO-text is not to be seen by default")
    captured = capsys.readouterr()
    assert "INFO-text" not in captured.out

    splitlogger.warning("This WARNING-text is to be seen in stdout")
    captured = capsys.readouterr()
    assert "WARNING-text" in captured.out
    assert "WARNING-text" not in captured.err

    splitlogger.error("This ERROR-text should only be in stderr")
    captured = capsys.readouterr()
    assert "ERROR-text" not in captured.out
    assert "ERROR-text" in captured.err

    # If output is written to stdout, all logs should go to stderr:
    nosplit_logger = pyscal.getLogger_pyscal(
        "test_levels_nosplit", args_dict={"output": "-", "debug": True}
    )
    nosplit_logger.debug("This DEBUG-text is to be seen in stderr")
    captured = capsys.readouterr()
    assert "DEBUG-text" not in captured.out
    assert "DEBUG-text" in captured.err

    nosplit_logger.info("This INFO-text is to be seen by in stderr")
    captured = capsys.readouterr()
    assert "INFO-text" not in captured.out
    assert "INFO-text" in captured.err

    nosplit_logger.warning("This WARNING-text is to be seen in stderr")
    captured = capsys.readouterr()
    assert "WARNING-text" not in captured.out
    assert "WARNING-text" in captured.err

    nosplit_logger.error("This ERROR-text should only be in stderr")
    captured = capsys.readouterr()
    assert "ERROR-text" not in captured.out
    assert "ERROR-text" in captured.err


def test_pyscal_logging_verbose(tmp_path, mocker, capsys):
    """Test that the command line client logs correctly with output set to
    stdout and verbose set to true.
    """

    testdir = Path(__file__).absolute().parent
    relperm_file = testdir / "data/relperm-input-example.xlsx"
    commands = ["pyscal", str(relperm_file), "--output", "-", "-v"]

    mocker.patch("sys.argv", commands)

    pyscalcli.main()
    captured = capsys.readouterr()
    stdout_output = captured.out
    stderr_output = captured.err

    assert "INFO:" in stderr_output
    assert "INFO:" not in stdout_output


def test_pyscal_logging(tmp_path, mocker, capsys):
    """Test that the command line client logs correctly with output set to
    stdout and verbose set to false.
    """

    testdir = Path(__file__).absolute().parent
    relperm_file = testdir / "data/relperm-input-example.xlsx"
    commands = ["pyscal", str(relperm_file), "--output", "-"]

    mocker.patch("sys.argv", commands)

    pyscalcli.main()
    captured = capsys.readouterr()
    stdout_output = captured.out
    stderr_output = captured.err

    assert "INFO:" not in stdout_output
    assert "INFO:" not in stderr_output


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires python 3.7 or higher")
def test_pyscal_logging_output_to_file(tmp_path, mocker, capsys):
    """Test that the command line client logs correctly with output set to
    file and verbose set to true.
    """

    testdir = Path(__file__).absolute().parent
    relperm_file = testdir / "data/relperm-input-example.xlsx"
    commands = ["pyscal", str(relperm_file), "--output", "relperm.inc", "-v"]

    result = subprocess.run(commands, cwd=tmp_path, capture_output=True, check=True)
    stdout_output = result.stdout.decode()
    stderr_output = result.stderr.decode()

    assert "INFO:" in stdout_output
    assert "INFO:" not in stderr_output


def test_repeated_logger_construction(capsys):
    """If we repeatedly call getLogger(), ensure handlers are not added on top"""
    logger = pyscal.getLogger_pyscal("nodouble")
    logger = pyscal.getLogger_pyscal("nodouble")
    logger.warning("Don't repeat me")
    captured = capsys.readouterr()
    assert captured.out.count("Don't repeat me") == 1



"""Container object for one WaterOil and one GasOil object"""

from typing import Optional

import numpy as np
import pandas as pd

import pyscal
from pyscal.constants import SWINTEGERS
from pyscal.utils.string import comment_formatter, df2str

from .gasoil import GasOil
from .wateroil import WaterOil

logger = pyscal.getLogger_pyscal(__name__)


class WaterOilGas:
    """A representation of three-phase properties for oil-water-gas

    Use one object for each satnum.

    One WaterOil and one GasOil object will be created, with compatible
    saturation ranges. Access the class members 'wateroil' and 'gasoil'
    directly to add curves.

    All arguments to the initializer can be defaulted, and all are zero except
    h.

    Either the gasoil or the wateroil member is allowed to be None in case of
    only two phases present. It is allowed for functions to set these members
    to None after initialization, as done in SCALrecommendation.interpolate()

    Args:
        swirr: Irreducible water saturation for capillary pressure
        swl: First water saturation point in outputted tables.
        swcr: Critical water saturation, water is immobile below this
        sorw: Residual oil saturation
        sgcr: Critical gas saturation, gas is immobile below this
        h: Saturation intervals in generated tables.
        tag: Optional text that will be included as comments.
        fast: Set to True if you prefer speed over robustness. Not recommended,
            pyscal will not guarantee valid output in this mode.
    """

    def __init__(
        self,
        swirr: float = 0.0,
        swl: float = 0.0,
        swcr: float = 0.0,
        sorw: float = 0.0,
        sorg: float = 0.0,
        sgcr: float = 0.0,
        h: Optional[float] = None,
        tag: str = "",
        fast: bool = False,
    ) -> None:
        """Sets up the saturation range for three phases"""
        self.fast: bool = fast
        self.wateroil: Optional[WaterOil] = WaterOil(
            swirr=swirr, swl=swl, swcr=swcr, sorw=sorw, h=h, tag=tag, fast=fast
        )
        self.gasoil: Optional[GasOil] = GasOil(
            swirr=swirr, sgcr=sgcr, sorg=sorg, swl=swl, h=h, tag=tag, fast=fast
        )

    def selfcheck(self) -> bool:
        """Run selfcheck on both wateroil and gasoil.

        Returns true only if both passes if both are present.

        If only wateroil or gasoil is present (the other
        being None, only the relevant is checked.
        """
        if self.wateroil is not None and self.gasoil is not None:
            return self.wateroil.selfcheck() and self.gasoil.selfcheck()
        if self.wateroil is not None:
            return self.wateroil.selfcheck()
        if self.gasoil is not None:
            return self.gasoil.selfcheck()
        logger.error("Both wateroil and gasoil are None in WaterOilGas")
        return False

    def SWOF(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Return a SWOF string. Delegated to the wateroil object"""
        if self.wateroil is None:
            logger.error("No WaterOil object in this WaterOilGas object")
            return ""
        if "KRW" not in self.wateroil.table or "KROW" not in self.wateroil.table:
            logger.error("Missing KRW/KROW curves in WaterOilGas object")
            return ""
        return self.wateroil.SWOF(header, dataincommentrow)

    def SGOF(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Return a SGOF string. Delegated to the gasoil object."""
        if self.gasoil is None:
            logger.error("No GasOil object in this WaterOilGas object")
            return ""
        if "KRG" not in self.gasoil.table or "KROG" not in self.gasoil.table:
            logger.error("Missing KRG/KROG curves in WaterOilGas object")
            return ""
        return self.gasoil.SGOF(header, dataincommentrow)

    def SLGOF(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Return a SLGOF string. Delegated to the gasoil object."""
        if self.gasoil is None:
            logger.error("No GasOil object in this WaterOilGas object")
            return ""
        if "KRG" not in self.gasoil.table or "KROG" not in self.gasoil.table:
            logger.error("Missing KRG/KROG in WaterOilGas object")
            return ""
        return self.gasoil.SLGOF(header, dataincommentrow)

    def SGFN(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Return a SGFN string. Delegated to the gasoil object."""
        if self.gasoil is None:
            logger.error("No GasOil object in this WaterOilGas object")
            return ""
        if "KRG" not in self.gasoil.table:
            logger.error("Missing KRG in WaterOilGas object")
            return ""
        return self.gasoil.SGFN(header, dataincommentrow)

    def SWFN(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Return a SWFN string. Delegated to the wateroil object."""
        if self.wateroil is None:
            logger.error("No WaterOil object in this WaterOilGas object")
            return ""
        if "KRW" not in self.wateroil.table:
            logger.error("Missing KRW in WaterOilGas object")
            return ""
        return self.wateroil.SWFN(header, dataincommentrow)

    def SOF3(self, header: bool = True, dataincommentrow: bool = True) -> str:
        """Return a SOF3 string, combining data from the wateroil and
        gasoil objects.

        So - the oil saturation ranges from 0 to 1-swl. The saturation points
        from the WaterOil object is used to generate these
        """
        if (self.wateroil is None or self.gasoil is None) or (
            "KROW" not in self.wateroil.table or "KROG" not in self.gasoil.table
        ):
            logger.error("Both WaterOil and GasOil krow/krog is needed for SOF3")
            return ""
        self.threephaseconsistency()

        # Copy of the wateroil data:
        table = pd.DataFrame(self.wateroil.table[["SW", "KROW"]])
        table["SO"] = 1 - table["SW"]

        # Copy of the gasoil data:
        gastable = pd.DataFrame(self.gasoil.table[["SG", "KROG"]])
        gastable["SO"] = 1 - gastable["SG"] - self.wateroil.swl

        # Merge WaterOil and GasOil on oil saturation, interpolate for
        # missing data (potentially different sg- and sw-grids)
        sof3table = (
            pd.concat([table, gastable], sort=True)
            .set_index("SO")
            .sort_index()
            .interpolate(method="slinear")
            .ffill()
            .bfill()
            .reset_index()
        )
        sof3table["soint"] = list(map(round, sof3table["SO"] * SWINTEGERS))
        sof3table = sof3table.drop_duplicates("soint")

        # The 'so' column has been calculated from floating point numbers
        # and the zero value easily becomes a negative zero, circumvent this:
        zerorow = np.isclose(sof3table["SO"], 0.0)
        sof3table.loc[zerorow, "SO"] = abs(sof3table.loc[zerorow, "SO"])

        string = ""
        if header:
            string += "SOF3\n"
        wo_tag = comment_formatter(self.wateroil.tag)
        go_tag = comment_formatter(self.gasoil.tag)
        if wo_tag != go_tag:
            string += wo_tag
            string += go_tag
        else:
            # Only print once if they are equal
            string += wo_tag
        string += "-- pyscal: " + str(pyscal.__version__) + "\n"
        if dataincommentrow:
            string += self.wateroil.swcomment
            string += self.gasoil.sgcomment
            string += self.wateroil.krowcomment
            string += self.gasoil.krogcomment

        width = 10
        string += (
            "-- "
            + "SO".ljust(width - 3)
            + "KROW".ljust(width)
            + "KROG".ljust(width)
            + "\n"
        )
        string += df2str(sof3table[["SO", "KROW", "KROG"]])
        string += "/\n"
        return string

    @property
    def swirr(self) -> float:
        """Get the swirr used for the WaterOil object"""
        if self.wateroil is None:
            raise ValueError("wateroil is None in WaterOilGas object")
        return self.wateroil.swirr

    @property
    def swl(self) -> float:
        """Get the swl used for the WaterOil object"""
        if self.wateroil is None:
            raise ValueError("wateroil is None in WaterOilGas object")
        return self.wateroil.swl

    @property
    def sorg(self) -> float:
        """Get the sorg used in the GasOil object"""
        if self.gasoil is None:
            raise ValueError("gasoil is None in WaterOilGas object")
        return self.gasoil.sorg

    @property
    def sorw(self) -> float:
        """Get the sorw used for the WaterOil object"""
        if self.wateroil is None:
            raise ValueError("wateroil is None in WaterOilGas object")
        return self.wateroil.sorw

    @property
    def tag(self) -> str:
        """Get the tag, a combination of the tags in WaterOil
        and GasOil only if they are different"""
        if self.wateroil is not None and self.gasoil is not None:
            if self.wateroil.tag == self.gasoil.tag:
                return self.wateroil.tag
            return self.wateroil.tag + " " + self.gasoil.tag
        if self.wateroil is not None:
            return self.wateroil.tag
        if self.gasoil is not None:
            return self.gasoil.tag
        return ""

    def threephaseconsistency(self) -> bool:
        """Perform consistency checks on produced curves, similar
        to what Eclipse does at startup.

        If any (likely) errors are found, these are printed
        as warnings and this function returns False.

        "Errors" from this function should be treated as
        warnings, as false positives from this function should
        not have breaking consequences.
        """

        # Eclipse errors:

        # 1: Error in saturation table number 1 at the maximum oil
        # saturation (0.9) krow and krog should both be equal the oil
        # relative permeability for a system with oil and connate
        # water only - but in this case they are different (krow=1.0
        # and krog=0.93)

        if self.wateroil is None or self.gasoil is None:
            # Nothing here to check if we only have two phases
            return True

        wog_is_ok = True
        if not np.isclose(
            self.wateroil.table["KROW"].max(), self.gasoil.table["KROG"].max()
        ):
            logger.warning(
                "Eclipse will fail, max(KROW)=%g is not equal to max(KROG)=%g",
                self.wateroil.table["KROW"].max(),
                self.gasoil.table["KROG"].max(),
            )
            wog_is_ok = False

        # 2: Inconsistent end points in saturation table 1 the maximum
        # gas saturation (0.91) plus the connate water saturation
        # (0.10) must not exceed 1.0
        if self.gasoil.table["SG"].max() + self.wateroil.table["SW"].min() > 1.0:
            logger.warning("Eclipse will fail, max(SG) + swl > 1.0")
            wog_is_ok = False

        # 3: Warning: Consistency problem for gas phase endpoint (krgr > krg)
        # in grid cell (i, j, k) for saturation end-points krgr=1.0
        # krg = 0.49.

        # 4: Warning: Consistency problem for oil phase endpoint (sgu > 1-swl)
        # in grid cell (i, j, k) for saturation endpoints sgu=0.78,
        # swl=0.45, (1-swl) = 0.55

        return wog_is_ok



"""Command line tool for pyscal"""

import argparse
import io
import sys
import traceback
from pathlib import Path
from typing import Optional

import pandas as pd

from pyscal import (
    GasWater,
    SCALrecommendation,
    WaterOilGas,
    __version__,
    getLogger_pyscal,
    plotting,
)
from pyscal.factory import (
    create_pyscal_list,
    create_scal_recommendation_list,
    load_relperm_df,
)

EPILOG = """
The parameter file should contain a table with at least the column
SATNUM, containing only consecutive integers starting at 1. Each row
provides the data for the corresponding SATNUM. Comments are put in a
column called TAG or COMMENT. Column headers are case insensitive.

Saturation endpoints are put in columns 'swirr', 'swl', 'swcr', 'sorw',
'sgcr' and 'sorg'. Relative permeability endpoints are put in columns
'krwend', 'krwmax', 'krowend', 'krogend', 'krgend' and 'krgmax'.
These columns are optional and are defaulted to 0 or 1.

Corey or LET parametrization are based on presence of the columns
'Nw', 'Now', 'Nog', 'Ng', 'Lw', 'Ew', 'Tw', 'Low', 'Eow', 'Tow',
'Log', 'Eog', 'Tog', 'Lg', 'Eg', 'Tg'.

Simple J-function for capillary pressure ("RMS" version) is used if the columns
'a', 'b', 'poro_ref', 'perm_ref' and 'drho' are found. If you provide
'a_petro', or 'b_petro', the petrophysical formulation of the simple J-function
is used. Check API for exact formulas. Normalized J-function is used if 'a',
'b', 'poro', 'perm' and 'sigma_costau' is provided.

Primary drainage and imbibition capillary pressure with LET parametrization are
based on presence of the columnns 'Lp', 'Ep', 'Tp','Lt', 'Et', 'Tt', 'Pcmax', 'Pct' and
'Ls', 'Es', 'Ts', 'Lf', 'Ef', 'Tf', 'Pcmax', 'Pct', 'Pcmin' respectively.

Capillary pressure from the Skjæveland correlation is based on the on the presence
of the columns 'cw', 'co', 'aw', 'ao'.

For SCAL recommendations, there should be exactly three rows for each SATNUM,
tagged with the strings 'low', 'base' and 'high' in the column 'CASE'

When interpolating in a SCAL recommendation, 'int_param_wo' is the main parameter
that is used for water-oil, gas-oil and gas-water, and for all SATNUMs if nothing
more is provided. Provide int_param_go in addition if separate interpolation
for WaterOil and GasOil is needed.
"""


def get_parser() -> argparse.ArgumentParser:
    """Construct the argparse parser for the command line script.

    Returns:
        argparse.Parser
    """
    parser = argparse.ArgumentParser(
        prog="pyscal",
        description=(
            "pyscal (" + __version__ + ") is a tool to create Eclipse include "
            "files for relative permeability input from tabulated parameters."
        ),
        epilog=EPILOG,
    )
    parser.add_argument(
        "parametertable",
        help=(
            "CSV or XLSX file with Corey or LET parameters for relperms. "
            "One SATNUM pr row."
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print informational messages while processing input",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Print debug information",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s (version " + __version__ + ")",
    )
    parser.add_argument(
        "-o",
        "--output",
        default="relperm.inc",
        help="Name of Eclipse include file to produce",
    )
    parser.add_argument(
        "--delta_s",
        default=None,
        type=float,
        help="Saturation table step-length for sw/sg. Default 0.01",
    )
    parser.add_argument(
        "--int_param_wo",
        default=None,
        type=float,
        help=(
            "Interpolation parameter for water-oil, needed if the parametertable "
            "contains pess/low, base and opt/high for each SATNUM. "
            "The parameter will be used for all SATNUMs and must be "
            "between -1 and 1. Also used for GasWater."
        ),
    )
    parser.add_argument(
        "--int_param_go",
        default=None,
        type=float,
        help=(
            "Interpolation parameter for gas-oil. "
            "If not provided, the water-oil interpolation parameter will be used "
            "as default. Do not use for GasWater."
        ),
    )
    parser.add_argument(
        "--sheet_name",
        type=str,
        default=None,
        help="Sheet name if reading XLSX file. Defaults to first sheet",
    )
    parser.add_argument(
        "--slgof",
        action="store_true",
        default=False,
        help="If using family 1 keywords, use SLGOF instead of SGOF",
    )
    parser.add_argument(
        "--family2",
        action="store_true",
        default=False,
        help=(
            "Output family 2 keywords, SWFN, SGFN and SOF3/SOF2. "
            "Family 1 (SWOF + SGOF) is written if this is not set.  "
            "Implicit for gas-water input."
        ),
    )

    # Plotting arguments
    parser.add_argument(
        "--plot",
        action="store_true",
        default=False,
        help=("Make and save relative permeability figures."),
    )
    parser.add_argument(
        "--plot_pc",
        action="store_true",
        default=False,
        help=("Make and save capillary pressure figures."),
    )
    parser.add_argument(
        "--plot_semilog",
        action="store_true",
        default=False,
        help=(
            "Plot relative permeability figures with log y-axis."
            "Run both with and without this flag to plot"
            "both linear and semi-log relperm plots."
        ),
    )
    parser.add_argument(
        "--plot_outdir",
        default="./",
        help="Directory where the plot output figures will be saved.",
    )

    return parser


def main() -> None:
    """Endpoint for pyscals command line utility.

    Translates from argparse API to Pyscal's Python API"""
    parser = get_parser()
    args = parser.parse_args()

    if isinstance(sys.stdout, io.TextIOWrapper):
        sys.stdout.reconfigure(encoding="UTF-8")
    if isinstance(sys.stdout, io.TextIOWrapper):
        sys.stdout.reconfigure(encoding="UTF-8")

    try:
        pyscal_main(
            parametertable=args.parametertable,
            verbose=args.verbose,
            debug=args.debug,
            output=args.output,
            delta_s=args.delta_s,
            int_param_wo=args.int_param_wo,
            int_param_go=args.int_param_go,
            sheet_name=args.sheet_name,
            slgof=args.slgof,
            family2=args.family2,
            plot=args.plot,
            plot_pc=args.plot_pc,
            plot_semilog=args.plot_semilog,
            plot_outdir=args.plot_outdir,
        )
    except (OSError, ValueError) as err:
        print("".join(traceback.format_tb(err.__traceback__)))
        sys.exit(str(err))


def pyscal_main(
    parametertable: str,
    verbose: bool = False,
    debug: bool = False,
    output: str = "relperm.inc",
    delta_s: Optional[float] = None,
    int_param_wo: Optional[float] = None,
    int_param_go: Optional[float] = None,
    sheet_name: Optional[str] = None,
    slgof: bool = False,
    family2: bool = False,
    plot: bool = False,
    plot_pc: bool = False,
    plot_semilog: bool = False,
    plot_outdir: str = "./",
) -> None:
    """A "main()" method not relying on argparse. This can be used
    for testing, and also by an ERT forward model, e.g.
    in semeio (github.com/equinor/semeio)

    Args:
        parametertable: Filename (CSV or XLSX) to load
        verbose: verbose or not
        debug: debug mode or not
        output: Output filename
        delta_s: Saturation step-length
        int_param_wo: Interpolation parameter for wateroil
        int_param_go: Interpolation parameter for gasoil
        sheet_name: Which sheet in XLSX file
        slgof: Use SLGOF
        family2: Dump family 2 keywords
        plot: Plot relative permeability figures and save
        plot_pc: Plot capillary pressure curves in addition to relperm curves
        plot_semilog: Plot relative permeability figures with log y-axis

    """

    logger = getLogger_pyscal(
        __name__, {"debug": debug, "verbose": verbose, "output": output}
    )

    parametertable = load_relperm_df(parametertable, sheet_name=sheet_name)

    assert isinstance(parametertable, pd.DataFrame)
    logger.debug("Input data:\n%s", parametertable.to_string(index=False))

    if int_param_go is not None and int_param_wo is None:
        raise ValueError("Don't use int_param_go alone, only int_param_wo")
    if isinstance(int_param_wo, list) or isinstance(int_param_go, list):
        raise TypeError(
            "SATNUM specific interpolation parameters are not supported in pyscalcli"
        )
    if int_param_wo is not None and "CASE" not in parametertable:
        raise ValueError(
            "Interpolation parameter provided but no CASE column in input data"
        )
    if "SATNUM" not in parametertable:
        raise ValueError("There is no column called SATNUM in the input data")

    if "CASE" in parametertable:
        # Then we should do interpolation
        if int_param_wo is None:
            raise ValueError("No interpolation parameters provided")
        scalrec_list = create_scal_recommendation_list(parametertable, h=delta_s)
        assert isinstance(scalrec_list[1], SCALrecommendation)
        if scalrec_list[1].type == WaterOilGas:
            logger.info(
                "Interpolating, wateroil=%s, gasoil=%s",
                str(int_param_wo),
                str(int_param_go),
            )
            wog_list = scalrec_list.interpolate(int_param_wo, int_param_go, h=delta_s)
        elif scalrec_list[1].type == GasWater:
            logger.info(
                "Interpolating, gaswater=%s",
                str(int_param_wo),
            )
            wog_list = scalrec_list.interpolate(int_param_wo, None, h=delta_s)
    else:
        wog_list = create_pyscal_list(
            parametertable, h=delta_s
        )  # can be both water-oil, water-oil-gas, or gas-water

    family = 2 if family2 or wog_list.pyscaltype == GasWater else 1

    if output == "-":
        print(wog_list.build_eclipse_data(family=family, slgof=slgof))
    else:
        if not Path(output).parent.exists():
            raise IOError(f"Output directory not found '{Path(output).parent}'")
        with open(output, "w", newline="\n", encoding="utf-8") as fh:
            fh.write(wog_list.build_eclipse_data(family=family, slgof=slgof))
        print("Written to " + output)

    if plot:
        plotting.plotter(wog_list, plot_pc, plot_semilog, plot_outdir)



"""Example code for benchmarking of the "fast" feature"""

import timeit

from pyscal import WaterOilGas


def benchme(fast=False, doprint=False):
    """Test function for benchmarking the "fast"-feature

    Pipe the following code snip into "ipython" on the shell:
    > echo "import benchme
      %timeit benchme.benchme()
      %timeit benchme.benchme(fast=True)
      " | ipython

    or run this module directly
    """
    wog = WaterOilGas(swl=0.1, h=0.1, fast=fast)
    wog.wateroil.add_corey_oil(now=3, kroend=0.9)
    wog.wateroil.add_corey_water(nw=2, krwend=0.2)
    wog.gasoil.add_corey_oil()
    wog.gasoil.add_corey_gas()
    if not doprint:
        len(wog.SWOF())
        len(wog.SGOF())
    else:
        print(wog.SWOF())
        print(wog.SGOF())


if __name__ == "__main__":
    print("Running in robust and slow mode:")
    print(
        timeit.timeit(
            stmt="benchme(fast=False)", setup="from benchme import benchme", number=100
        )
    )
    print("Running in fast mode:")
    print(
        timeit.timeit(
            stmt="benchme(fast=True)", setup="from benchme import benchme", number=100
        )
    )
